# -*- coding: utf-8 -*-
from stat import S_ISDIR,S_ISREG
import StringIO
import paramiko
import os
import glob

import aiida.transport
from aiida.common.utils import escape_for_bash
from aiida.transport import FileAttribute
from aiida.common import aiidalogger

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

# TODO : callback functions in paramiko are currently not used much and probably broken
def parse_sshconfig(computername):
    config = paramiko.SSHConfig()
    try:
        config.parse(open(os.path.expanduser('~/.ssh/config')))
    except IOError:
        # No file found, so empty configuration
        pass
    
    return config.lookup(computername)

def convert_to_bool(string):
    upstring = str(string).upper()
    
    if upstring in ['Y', 'YES', 'T', 'TRUE']:
        return True
    elif upstring in ['N', 'NO', 'F', 'FALSE']:
        return False
    else:
        raise ValueError("Invalid boolean value provided")

class SshTransport(aiida.transport.Transport):
    """
    Support connection, command execution and data transfer to remote computers via SSH+SFTP.
    """
    # Valid keywords accepted by the connect method of paramiko.SSHClient
    # I disable 'password' and 'pkey' to avoid these data to get logged in the
    # aiida log file.
    _valid_connect_params = ['username', 'port', 'look_for_keys', 
                             'key_filename', 'timeout', 'allow_agent',
                             'proxy_command', # Managed 'manually' in connect
                             'compress']
    
    # Valid parameters for the ssh transport
    # For each param, a class method with name
    # _convert_PARAMNAME_fromstring
    # should be defined, that returns the value converted from a string to
    # a correct type, or raise a ValidationError 
    #
    # moreover, if you want to help in the default configuration, you can 
    # define a _get_PARAMNAME_suggestion_string
    # to return a suggestion; it must accept only one parameter, being a Computer
    # instance
    _valid_auth_params = _valid_connect_params + [
        'load_system_host_keys',
        'key_policy',
        ]
    
    @classmethod
    def _convert_username_fromstring(cls, string):
        """
        Convert the username from string.
        """
        return string

    @classmethod
    def _get_username_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        import getpass
        
        config = parse_sshconfig(computer.hostname)
        # Either the configured user in the .ssh/config, or the current username
        return str(config.get('user',getpass.getuser())) 
    

    @classmethod
    def _convert_port_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        try:
            return int(string)
        except ValueError:
            raise ValidationError("The port must be an integer")

    @classmethod
    def _get_port_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        config = parse_sshconfig(computer.hostname)
        # Either the configured user in the .ssh/config, or the default SSH port
        return str(config.get('port',22))

    @classmethod
    def _convert_key_filename_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        path = os.path.expanduser(string)
        
        if not os.path.isabs(path):
            raise ValidationError("The key filename must be an absolute path")

        if not os.path.exists(path):
            raise ValidationError("The key filename must exist")

        return path
    
    @classmethod
    def _get_key_filename_suggestion_string(cls, computer):        
        """
        Return a suggestion for the specific field.
        """
        config = parse_sshconfig(computer.hostname)
        
        try:
            identities = config['identityfile']
            # In paramiko > 0.10, identity file is a list of strings.
            if isinstance(identities,basestring):
                identity = identities
            elif isinstance(identities,(list,tuple)):
                if not identities:
                    # An empty list should not be provided; to be sure,
                    # anyway, behave as if no identityfile were defined
                    raise KeyError
                # By default we suggest only the first one
                identity = identities[0]
            else:
                # If the parser provides an unknown type, just skip to
                # the 'except KeyError' section, as if no identityfile 
                # were provided (hopefully, this should never happen)
                raise KeyError
        except KeyError:
            # No IdentityFile defined: return an empty string
            return ""
        
        return os.path.expanduser(identity)

    @classmethod
    def _convert_timeout_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        try:
            return int(string)
        except ValueError:
            raise ValidationError("The timeout must be an integer")

    @classmethod
    def _get_timeout_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        
        Provide 60s as a default timeout for connections.
        """
        config = parse_sshconfig(computer.hostname)
        return str(config.get('connecttimeout',"60"))

    @classmethod
    def _convert_allow_agent_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        try:
            return convert_to_bool(string)
        except ValueError:
            raise ValidationError("Allow_agent must be an boolean")

    @classmethod
    def _get_allow_agent_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        return ""

    @classmethod
    def _convert_look_for_keys_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        try:
            return convert_to_bool(string)
        except ValueError:
            raise ValidationError("look_for_keys must be an boolean")

    @classmethod
    def _get_look_for_keys_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        return ""

    @classmethod
    def _convert_proxy_command_fromstring(cls, string):
        """
        Convert the proxy command from string.
        """
        from aiida.common.exceptions import ValidationError
        
        if str(string).strip():
            return str(string)
        else:
            return None

    @classmethod
    def _get_proxy_command_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        config = parse_sshconfig(computer.hostname)
        # Either the configured user in the .ssh/config, or the default SSH port
        raw_string = str(config.get('proxycommand',''))
        # Note: %h and %p get already automatically substituted with 
        # hostname and port by the config parser!
        
        pieces = raw_string.split()
        new_pieces = []
        for piece in pieces:
            if '>' in piece:
                # If there is a piece with > to readdress stderr or stdout,
                # skip from here on (anything else can only be readdressing)
                break
            else:
                new_pieces.append(piece)
        return " ".join(new_pieces)

    @classmethod
    def _convert_compress_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        try:
            return convert_to_bool(string)
        except ValueError:
            raise ValidationError("compress must be an boolean")

    @classmethod
    def _get_compress_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        return "True"
    
    @classmethod
    def _convert_load_system_host_keys_fromstring(cls, string):
        """
        Convert the port from string.
        """
        from aiida.common.exceptions import ValidationError
        
        try:
            return convert_to_bool(string)
        except ValueError:
            raise ValidationError("compress must be an boolean")

    @classmethod
    def _get_load_system_host_keys_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        return "True"
    
    @classmethod
    def _convert_key_policy_fromstring(cls, string):
        """
        Convert the port from string.
        """
        return string
    
    @classmethod
    def _get_key_policy_suggestion_string(cls, computer):
        """
        Return a suggestion for the specific field.
        """
        return "RejectPolicy"
    
        
    def __init__(self, machine, **kwargs):
        """
        Initialize the SshTransport class.
        
        :param machine: the machine to connect to
        :param load_system_host_keys: (optional, default False): if False, do not
                load the system host keys
        :param key_policy: (optional, default = paramiko.RejectPolicy()): the
                policy to use for unknown keys
        Other parameters valid for the ssh connect function (see the 
        self._valid_connect_params list) are passed to the connect
        function (as port, username, password, ...); taken from the
        accepted paramiko.SSHClient.connect() params)
        """
        super(SshTransport,self).__init__()

        self._is_open = False
        self._sftp = None
        
        self._machine = machine

        self._client = paramiko.SSHClient()
        self._load_system_host_keys = kwargs.pop(
            'load_system_host_keys', False)
        if self._load_system_host_keys:
            self._client.load_system_host_keys()

        self._missing_key_policy = kwargs.pop(
            'key_policy','RejectPolicy') # This is paramiko default
        if self._missing_key_policy == 'RejectPolicy':
            self._client.set_missing_host_key_policy(paramiko.RejectPolicy())
        elif self._missing_key_policy == 'WarningPolicy':
            self._client.set_missing_host_key_policy(paramiko.WarningPolicy())
        elif self._missing_key_policy == 'AutoAddPolicy':
            self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        else:
            raise ValueError("Unknown value of the key policy, allowed values "
                             "are: RejectPolicy, WarningPolicy, AutoAddPolicy")

        self._connect_args = {}
        for k in self._valid_connect_params:
            try:
                self._connect_args[k] = kwargs.pop(k)
            except KeyError:
                pass

        if kwargs:
            raise ValueError("The following parameters were not accepted by "
                             "the transport: {}".format(
                    ",".join(str(k) for k in kwargs)))

    def open(self):
        """
        Open a SSHClient to the machine possibly using the parameters given
        in the __init__. 
        
        Also opens a sftp channel, ready to be used.
        The current working directory is set explicitly, so it is not None.
        
        :raise InvalidOperation: if the channel is already open
        """
        from aiida.common.exceptions import InvalidOperation
        
        if self._is_open:
            raise InvalidOperation("Cannot open the transport twice")
        # Open a SSHClient
        connection_arguments = self._connect_args
        proxystring = connection_arguments.pop('proxy_command', None)
        if proxystring is not None:
            proxy = paramiko.ProxyCommand(proxystring)
            connection_arguments['sock'] = proxy
            
        try:
            self._client.connect(self._machine,**connection_arguments)
        except Exception as e:
            self.logger.error("Error connecting through SSH: [{}] {}, "
                              "connect_args were: {}".format(
                                  e.__class__.__name__,
                                  e.message,
                                  self._connect_args))
            raise

        # Open also a SFTPClient
        self._sftp = self._client.open_sftp()
        # Set the current directory to a explicit path, and not to None
        self._sftp.chdir(self._sftp.normalize('.'))

        self._is_open = True
        
        return self
        
    def close(self):
        """
        Close the SFTP channel, and the SSHClient.
        
        :todo: correctly manage exceptions
        
        :raise InvalidOperation: if the channel is already open
        """
        from aiida.common.exceptions import InvalidOperation
        
        if not self._is_open:
            raise InvalidOperation("Cannot close the transport: "
                                         "it is already closed")

        self._sftp.close()
        self._client.close()
        self._is_open = False

    @property
    def sshclient(self):
        if not self._is_open:
            raise aiida.transport.TransportInternalError(
                "Error, ssh method called for SshTransport "
                "without opening the channel first")
        return self._client

    @property
    def sftp(self):
        if not self._is_open:
            raise aiida.transport.TransportInternalError(
                "Error, sftp method called for SshTransport "
                "without opening the channel first")
        return self._sftp

    def __str__(self):
        """
        Return a useful string.
        """
        conn_info = self._machine
        try:
            conn_info = "{}@{}".format(self._connect_args['username'],
                                       conn_info)
        except KeyError:
            # No username explicitly defined: ignore
            pass
        try:
            conn_info += ':{}'.format(self._connect_args['port'])
        except KeyError:
        # No port explicitly defined: ignore
            pass
            
        return "{} [{}]".format("OPEN" if self._is_open else "CLOSED",
                                conn_info)

    def chdir(self, path):
        """
        Change directory of the SFTP session. Emulated internally by
        paramiko. 

        Differently from paramiko, if you pass None to chdir, nothing
        happens and the cwd is unchanged.
        """
        old_path = self.sftp.getcwd()
        if path is not None:
            self.sftp.chdir(path)
        
        # Paramiko already checked that path is a folder, otherwise I would
        # have gotten an exception. Now, I want to check that I have read
        # permissions in this folder (nothing is said on write permissions,
        # though).
        # Otherwise, if I do _exec_command_internal, that as a first operation 
        # cd's in a folder, I get a wrong retval, that is an unwanted behavior.
        #
        # Note: I don't store the result of the function; if I have no
        # read permissions, this will raise an exception.
        try:
            self.sftp.stat('.')
        except IOError as e:
            if 'Permission denied' in e:
                self.chdir(old_path)
            raise IOError(e)


    def normalize(self, path):
        """
        Returns the normalized path (removing double slashes, etc...)
        """
        return self.sftp.normalize(path)
        
    def getcwd(self):
        """
        Return the current working directory for this SFTP session, as
        emulated by paramiko. If no directory has been set with chdir,
        this method will return None. But in __enter__ this is set explicitly,
        so this should never happen within this class.
        """
        return self.sftp.getcwd()


    def makedirs(self,path,ignore_existing=False):
        """
        Super-mkdir; create a leaf directory and all intermediate ones.
        Works like mkdir, except that any intermediate path segment (not
        just the rightmost) will be created if it does not exist.

        NOTE: since os.path.split uses the separators as the host system
        (that could be windows), I assume the remote computer is Linux-based
        and use '/' as separators!

        :param (str) path: directory to create
        :param (bool) ignore_existing: if set to true, it doesn't give any error
            if the leaf directory does already exist

        :raise OSError: If the directory already exists.
        """
        # check to avoid creation of empty dirs
        path = os.path.normpath(path)
        
        if path.startswith('/'):
            to_create = path.strip().split('/')[1:]
            this_dir='/'
        else:
            to_create = path.strip().split('/')
            this_dir = ''
            
        for count,element in enumerate(to_create):
            if count>0:
                this_dir += '/'
            this_dir += element
            if count+1==len(to_create) and self.isdir(this_dir) and ignore_existing:
                return
            if count+1==len(to_create) and self.isdir(this_dir) and not ignore_existing:
                self.mkdir(this_dir)
            if not self.isdir(this_dir):
                self.mkdir(this_dir)


    
    def mkdir(self,path,ignore_existing=False):
        """
        Create a folder (directory) named path.

        :param (str) path: name of the folder to create
        :param (bool) ignore_existing: if True, does not give any error if the directory
                  already exists

        :raise OSError: If the directory already exists.
        """
        if ignore_existing and self.isdir(path):
            return
        
        try:
            self.sftp.mkdir(path)
        except IOError as e:
            if os.path.isabs(path):
                raise OSError(
                    "Error during mkdir of '{}', "
                    "maybe you don't have the permissions to do it, "
                    "or the directory already exists? ({})".format(
                        path, e.message))
            else:
                raise OSError(
                    "Error during mkdir of '{}' from folder '{}', "
                    "maybe you don't have the permissions to do it, "
                    "or the directory already exists? ({})".format(
                        path, self.getcwd(), e.message))

    # TODO : implement rmtree
    def rmtree(self,path):
        """
        Remove a file or a directory at path, recursively
        Flags used: -r: recursive copy; -f: force, makes the command non interactive;

        :param (str) path: remote path to delete

        :raise IOError: if the rm execution failed.
        """
        # Assuming linux rm command!
        
        # TODO : do we need to avoid the aliases when calling rm_exe='rm'? Call directly /bin/rm?
                
        rm_exe = 'rm'
        rm_flags = '-r -f'
        # if in input I give an invalid object raise ValueError
        if not path:
            raise ValueError('Input to rmtree() must be a non empty string. ' +
                             'Found instead %s as path' % path)
        
        command = '{} {} {}'.format(rm_exe,
                                       rm_flags,
                                       escape_for_bash(path))

        retval,stdout,stderr = self.exec_command_wait(command)
        
        if retval == 0:
            if stderr.strip():
                self.logger.warning("There was nonempty stderr in the rm "
                                    "command: {}".format(stderr))
            return True
        else:
            self.logger.error("Problem executing rm. Exit code: {}, stdout: '{}', "
                              "stderr: '{}'".format(retval, stdout, stderr))
            raise IOError("Error while executing rm. Exit code: {}".format(retval) )

    
    def rmdir(self, path):
        """
        Remove the folder named 'path' if empty.
        """
        self.sftp.rmdir(path)


    def isdir(self,path):
        """
        Return True if the given path is a directory, False otherwise.
        Return False also if the path does not exist.
        """
        # Return False on empty string (paramiko would map this to the local
        # folder instead)
        if not path:
            return False
        try:
            return S_ISDIR(self.sftp.stat(path).st_mode)
        except IOError as e:
            if getattr(e,"errno",None) == 2:
                # errno=2 means path does not exist: I return False
                return False
            else:
                raise # Typically if I don't have permissions (errno=13)


    def chmod(self,path,mode):
        """
        Change permissions to path

        :param (str) path: path to file
        :param (int) mode: new permission bits
        """
        if not path:
            raise IOError("Input path is an empty argument.")
        return self.sftp.chmod(path,mode)


    def _os_path_split_asunder(self, path):
        """
        Used by makedirs. Takes path (a str)
        and returns a list deconcatenating the path
        """
        parts = []
        while True:
            newpath, tail = os.path.split(path)
            if newpath == path:
                assert not tail
                if path: parts.append(path)
                break
            parts.append(tail)
            path = newpath
        parts.reverse()
        return parts
            

    def put(self,localpath,remotepath,callback=None,dereference=True,overwrite=True,
            ignore_nonexisting=False):
        """
        Put a file or a folder from local to remote.
        redirects to putfile or puttree.
        :param (str) localpath: an (absolute) local path
        :param (str) remotepath: a remote path
        :param (bool) dereference: follow symbolic links.
            Default = True (default behaviour in paramiko). False is not implemented.
        :param (bool) overwrite: if True overwrites files and folders.
            Default = False

        :raise ValueError: if local path is invalid
        :raise OSError: if the localpath does not exist
        """
        # TODO: flag confirm exists since v1.7.7. What is the paramiko
        # version supported?

        # TODO : add dereference
        if not dereference:
            raise NotImplementedError

        if not os.path.isabs(localpath):
            raise ValueError("The localpath must be an absolute path")

        if self.has_magic(localpath):
            if self.has_magic(remotepath):
                raise ValueError("Pathname patterns are not allowed in the "
                                 "destination")

            # use the imported glob to analyze the path locally
            to_copy_list = glob.glob(localpath)
            
            rename_remote = False
            if len(to_copy_list)>1:
                # I can't scp more than one file on a single file
                if self.isfile(remotepath):
                    raise OSError("Remote destination is not a directory")
                # I can't scp more than one file in a non existing directory
                elif not self.path_exists(remotepath):  # questo dovrebbe valere solo per file
                    raise OSError("Remote directory does not exist")
                else: # the remote path is a directory
                    rename_remote=True

            for s in to_copy_list:
                if os.path.isfile(s):
                    if rename_remote: # copying more than one file in one directory 
                        # here is the case isfile and more than one file
                        r = os.path.join(remotepath,os.path.split(s)[1])
                        self.putfile(s,r,callback,dereference,overwrite )
                        
                    elif self.isdir(remotepath): # one file to copy in '.'
                        r = os.path.join(remotepath,os.path.split(s)[1])
                        self.putfile(s,r,callback,dereference,overwrite )
                    else: # one file to copy on one file
                        self.putfile(s,remotepath,callback,dereference,overwrite )
                else:
                    self.puttree(s,remotepath,callback,dereference,overwrite )

        else:
            if os.path.isdir(localpath):
                self.puttree( localpath,remotepath,callback,dereference,overwrite )
            elif os.path.isfile(localpath):
                if self.isdir(remotepath):
                    r = os.path.join(remotepath,os.path.split(localpath)[1])
                    self.putfile( localpath,r,callback,dereference,overwrite )
                else:
                    self.putfile( localpath,remotepath,callback,dereference,overwrite )
            else:
                if ignore_nonexisting:
                    pass
                else:
                    raise OSError("The local path {} does not exist"
                                  .format(localpath))


    def putfile(self,localpath,remotepath,callback=None,dereference=True,overwrite=True):
        """
        Put a file from local to remote.

        :param (str) localpath: an (absolute) local path
        :param (str) remotepath: a remote path
        :param (bool) overwrite: if True overwrites files and folders. 
            Default = True.

        :raise ValueError: if local path is invalid
        :raise OSError: if the localpath does not exist,
                    or unintentionally overwriting
        """
        # TODO : add dereference
        if not dereference:
            raise NotImplementedError

        # TODO : check what happens if I give in input a directory
        
        if not os.path.isabs(localpath):
            raise ValueError("The localpath must be an absolute path")

        if self.isfile(remotepath) and not overwrite:
            raise OSError('Destination already exists: not overwriting it')
        
        return self.sftp.put(localpath,remotepath,callback=callback)


    def puttree(self,localpath,remotepath,callback=None,dereference=True,overwrite=True): # by default overwrite
        """
        Put a folder recursively from local to remote.

        :param (str) localpath: an (absolute) local path
        :param (str) remotepath: a remote path
        :param (bool) dereference: follow symbolic links.
            Default = True (default behaviour in paramiko). False is not implemented.
        :param (bool) overwrite: if True overwrites files and folders.
            Default = True

        :raise ValueError: if local path is invalid
        :raise OSError: if the localpath does not exist, or trying to overwrite
        :raise IOError: if remotepath is invalid

        .. note:: setting dereference equal to True could cause infinite loops.
              see os.walk() documentation
        """
        # TODO : add dereference
        if not dereference:
            raise NotImplementedError

        if not os.path.isabs(localpath):
            raise ValueError("The localpath must be an absolute path")
        
        if not os.path.exists(localpath):
            raise OSError("The localpath does not exists")

        if not os.path.isdir(localpath):
            raise ValueError("Input localpath is not a folder: {}".format(localpath))

        if not remotepath:
            raise IOError("remotepath must be a non empty string")

        if self.path_exists(remotepath) and not overwrite:
            raise OSError("Can't overwrite existing files")
        if self.isfile(remotepath):
            raise OSError("Cannot copy a directory into a file")
        
        if not self.isdir(remotepath):  # in this case copy things in the remotepath directly
            self.mkdir(remotepath) # and make a directory at its place
        else: # remotepath exists already: copy the folder inside of it!
            remotepath = os.path.join(remotepath,os.path.split(localpath)[1])
            self.mkdir(remotepath) # create a nested folder
        
        # TODO, NOTE: we are not using 'onerror' because we checked above that
        # the folder exists, but it would be better to use it
        for this_source in os.walk(localpath):
            # Get the relative path
            this_basename = os.path.relpath(path=this_source[0],
                                            start=localpath)
            
            try:
                self.sftp.stat( os.path.join(remotepath,this_basename) )
            except IOError as e:
                import errno
                if e.errno == errno.ENOENT: # Missing file
                    self.mkdir( os.path.join(remotepath,this_basename) )
                else:
                    raise
                
            for this_file in this_source[2]:
                this_local_file = os.path.join(localpath,this_basename,this_file)        
                this_remote_file = os.path.join(remotepath,this_basename,this_file)
                self.putfile(this_local_file,this_remote_file)


    def get(self,remotepath,localpath,callback=None,dereference=True,overwrite=True,
            ignore_nonexisting=False):
        """
        Get a file or folder from remote to local.
        Redirects to getfile or gettree.

        :param (str) remotepath: a remote path
        :param (str) localpath: an (absolute) local path
        :param (bool) dereference: follow symbolic links.
            Default = True (default behaviour in paramiko). 
            False is not implemented.
        :param (bool) overwrite: if True overwrites files and folders.
            Default = False

        :raise ValueError: if local path is invalid
        :raise IOError: if the remotepath is not found
        """
        # TODO : add dereference
        if not dereference:
            raise NotImplementedError

        if not os.path.isabs(localpath):
            raise ValueError("The localpath must be an absolute path")

        if self.has_magic(remotepath):
            if self.has_magic(localpath):
                raise ValueError("Pathname patterns are not allowed in the "
                                 "destination")
            # use the self glob to analyze the path remotely
            to_copy_list = self.glob(remotepath)
            
            rename_local = False
            if len(to_copy_list)>1:
                # I can't scp more than one file on a single file
                if os.path.isfile(localpath):
                    raise IOError("Remote destination is not a directory")
                # I can't scp more than one file in a non existing directory
                elif not os.path.exists(localpath):  # this should hold only for files
                    raise OSError("Remote directory does not exist")
                else: # the remote path is a directory
                    rename_local=True
      
            for s in to_copy_list:
                if self.isfile(s):
                    if rename_local: # copying more than one file in one directory 
                        # here is the case isfile and more than one file
                        r = os.path.join(localpath,os.path.split(s)[1])
                        self.getfile(s,r,callback,dereference,overwrite )
                    else: # one file to copy on one file
                        self.getfile(s,localpath,callback,dereference,overwrite )
                else:
                    self.gettree(s,localpath,callback,dereference,overwrite )

        else:
            if self.isdir(remotepath):
                self.gettree( remotepath,localpath,callback,dereference,overwrite )
            elif self.isfile(remotepath):
                if os.path.isdir(localpath):
                    r = os.path.join(localpath,os.path.split(remotepath)[1])
                    self.getfile( remotepath,r,callback,dereference,overwrite )
                else:
                    self.getfile( remotepath,localpath,callback,dereference,overwrite )
            else:
                if ignore_nonexisting:
                    pass
                else:
                    raise IOError("The remote path {} does not exist"
                                  .format(remotepath))

    def getfile(self,remotepath,localpath,callback=None,dereference=True,overwrite=True):
        """
        Get a file from remote to local.

        :param (str) remotepath: a remote path
        :param (str) localpath: an (absolute) local path
        :param (bool) overwrite: if True overwrites files and folders.
                Default = False

        :raise ValueError: if local path is invalid
        :raise OSError: if unintentionally overwriting
        """
        # TODO : add dereference

        if not os.path.isabs(localpath):
            raise ValueError("localpath must be an absolute path")

        if os.path.isfile(localpath) and not overwrite:            
            raise OSError('Destination already exists: not overwriting it')

        if not dereference:
            raise NotImplementedError
            
        return self.sftp.get(remotepath,localpath,callback)
        
        
    def gettree(self,remotepath,localpath,callback=None,dereference=True,overwrite=True):
        """
        Get a folder recursively from remote to local.

        :param (str) remotepath: a remote path
        :param (str) localpath: an (absolute) local path
        :param (bool) dereference: follow symbolic links. 
            Default = True (default behaviour in paramiko). 
            False is not implemented.
        :param (bool) overwrite: if True overwrites files and folders.
            Default = False

        :raise ValueError: if local path is invalid
        :raise IOError: if the remotepath is not found
        :raise OSError: if unintentionally overwriting
        """
        # TODO : add dereference
        if not dereference:
            raise NotImplementedError

        if not remotepath:
            raise IOError("Remotepath must be a non empty string")
        if not localpath:
            raise ValueError("Localpaths must be a non empty string")

        if not os.path.isabs(localpath):
            raise ValueError("Localpaths must be an absolute path")

        if not self.isdir(remotepath):
            raise IOError("Input remotepath is not a folder: {}".format(localpath))

        if os.path.exists(localpath) and not overwrite:
            raise OSError("Can't overwrite existing files")
        if os.path.isfile(localpath):
            raise OSError("Cannot copy a directory into a file")
        
        if not os.path.isdir(localpath):  # in this case copy things in the remotepath directly
            os.mkdir(localpath) # and make a directory at its place
        else: # localpath exists already: copy the folder inside of it!
            localpath = os.path.join(localpath,os.path.split(remotepath)[1])
            os.mkdir(localpath) # create a nested folder
        
        item_list = self.listdir(remotepath)
        dest = str(localpath)

        for item in item_list:
            item = str(item)

            if self.isdir( os.path.join(remotepath,item) ):
                self.gettree( os.path.join(remotepath,item) , os.path.join(dest,item) )
            else:
                self.getfile( os.path.join(remotepath,item) , os.path.join(dest,item) )


    def get_attribute(self,path):
        """
        Returns the object Fileattribute, specified in aiida.transport
        Receives in input the path of a given file.
        """
        paramiko_attr = self.sftp.lstat(path)
        aiida_attr = FileAttribute()
        # map the paramiko class into the aiida one
        # note that paramiko object contains more informations than the aiida
        for key in aiida_attr._valid_fields:
            aiida_attr[key] = getattr(paramiko_attr,key)
        return aiida_attr


    def copyfile(self,remotesource,remotedestination,dereference=False,
                 pattern=None):
        """
        Copy a file from remote source to remote destination
        Redirects to copy().
        
        :param (str) remotesource:
        :param (str) remotedestination:
        :param (bool) dereference:
        :param (str) pattern:
        """
        cp_flags = '-f'
        return self.copy(remotesource,remotedestination,dereference,cp_flags,pattern)


    def copytree(self,remotesource,remotedestination,dereference=False,
                 pattern=None):
        """
        copy a folder recursively from remote source to remote destination
        Redirects to copy()

        :param (str) remotesource:
        :param (str) remotedestination:
        :param (bool) dereference:
        :param (str) pattern:
        """
        cp_flags = '-r -f'
        return self.copy(remotesource,remotedestination,dereference,cp_flags,pattern)


    def copy(self,remotesource,remotedestination,dereference=False):
        """
        Copy a file or a directory from remote source to remote destination.
        Flags used: -r: recursive copy; -f: force, makes the command non interactive;
          -L follows symbolic links

        :param (str) remotesource: file to copy from
        :param (str) remotedestination: file to copy to
        :param (bool) dereference: if True, copy content instead of copying the symlinks only
            Default = False.
        :raise IOError: if the cp execution failed.

        .. note:: setting dereference equal to True could cause infinite loops.
        """
        # In the majority of cases, we should deal with linux cp commands
        
        # TODO : do we need to avoid the aliases when calling cp_exe='cp'? Call directly /bin/cp?
        
        # TODO: verify that it does not re
        
        # For the moment, these are hardcoded. They may become parameters
        # as soon as we see the need.
        cp_flags='-r -f'
        cp_exe='cp'
        
        ## To evaluate if we also want -p: preserves mode,ownership and timestamp
        if dereference:
            # use -L; --dereference is not supported on mac
            cp_flags+=' -L'

        # if in input I give an invalid object raise ValueError
        if not remotesource:
            raise ValueError('Input to copy() must be a non empty string. ' +
                             'Found instead %s as remotesource' % remotesource)

        if not remotedestination:
            raise ValueError('Input to copy() must be a non empty string. ' +
                             'Found instead %s as remotedestination' % remotedestination)
        
        if self.has_magic(remotedestination):
            raise ValueError("Pathname patterns are not allowed in the "
                             "destination")

        if self.has_magic(remotesource):
            to_copy_list = self.glob(remotesource)
            
            if len(to_copy_list)>1:
                if not self.path_exists(remotedestination) or self.isfile(remotedestination): 
                    raise OSError("Can't copy more than one file in the same "
                                  "destination file")
            
            for s in to_copy_list:
                self._exec_cp(cp_exe,cp_flags,s,remotedestination)

        else:
            self._exec_cp(cp_exe,cp_flags,remotesource,remotedestination)


    def _exec_cp(self,cp_exe,cp_flags,src,dst):
        # to simplify writing the above copy function
        command = '{} {} {} {}'.format(cp_exe,cp_flags,escape_for_bash(src),
                                       escape_for_bash(dst))

        retval,stdout,stderr = self.exec_command_wait(command)
        
        # TODO : check and fix below
            
        if retval == 0:
            if stderr.strip():
                self.logger.warning("There was nonempty stderr in the cp "
                                        "command: {}".format(stderr))
        else:
            self.logger.error("Problem executing cp. Exit code: {}, stdout: '{}', "
                              "stderr: '{}', command: '{}'"
                              .format(retval, stdout, stderr,command))
            raise IOError("Error while executing cp. Exit code: {}, "
                          "stdout: '{}', stderr: '{}', "
                          "command: '{}'".format(retval, stdout, stderr,
                                                 command) )

                
    def _local_listdir(self,path,pattern=None):
        # acts on the local folder, for the rest, same as listdir
        if not pattern:
            return os.listdir( path )
        else:
            import re
            if path.startswith('/'): # always this is the case in the local case
                base_dir = path
            else:
                base_dir = os.path.join(os.getcwd(),path)

            filtered_list = glob.glob(os.path.join(base_dir,pattern) )
            if not base_dir.endswith(os.sep):
                base_dir+=os.sep
            return [ re.sub(base_dir,'',i) for i in filtered_list ]


    def listdir(self,path='.',pattern=None):
        """
        Get the list of files at path.
        :param (str) path: default = '.'
        :param (str) filter: returns the list of files matching pattern.
                             Unix only. (Use to emulate ls * for example)
        """
        if not pattern:
            return self.sftp.listdir(path)
        else:
            import re
            if path.startswith('/'):
                base_dir = path
            else:
                base_dir = os.path.join(self.getcwd(),path)

            filtered_list = glob.glob(os.path.join(base_dir,pattern) )
            if not base_dir.endswith('/'):
                base_dir += '/'
            return [ re.sub(base_dir,'',i) for i in filtered_list ]


    def remove(self,path):
        """
        Remove a single file at 'path'
        """
        return self.sftp.remove(path)

    def rename(self,src,dst):
        """
        Rename a file or folder from src to dst.

        :param str oldpath: existing name of the file or folder
        :param str newpath: new name for the file or folder

        :raises IOError: if src/dst is not found
        :raises ValueError: if src/dst is not a valid string
        """
        if not src:
            raise ValueError("Source {} is not a valid string".format(src))
        if not dst:
            raise ValueError("Destination {} is not a valid string".format(dst))
        if not self.isfile( src ):
            if not self.isdir( src ):
                raise IOError("Source {} does not exist".format(src))
        if not self.isfile( dst ):
            if not self.isdir( dst ):
                raise IOError("Destination {} does not exist".format(dst))

        return self.sftp.rename(src,dst)
    
    def isfile(self,path):
        """
        Return True if the given path is a file, False otherwise.
        Return False also if the path does not exist.
        """
        # This should not be needed for files, since an empty string should
        # be mapped by paramiko to the local directory - which is not a file -
        # but this is just to be sure
        if not path:
            return False
        try:
            self.logger.debug("stat for path '{}' ('{}'): {} [{}]".format(
                    path, self.sftp.normalize(path),
                    self.sftp.stat(path),self.sftp.stat(path).st_mode))
            return S_ISREG(self.sftp.stat(path).st_mode)
        except IOError as e:
            if getattr(e,"errno",None) == 2:
                # errno=2 means path does not exist: I return False
                return False
            else:
                raise # Typically if I don't have permissions (errno=13)

    def _exec_command_internal(self,command,combine_stderr=False,bufsize=-1):
        """
        Executes the specified command, first changing directory to the
        current working directory are returned by
        self.getcwd().
        Does not wait for the calculation to finish.

        For a higher-level _exec_command_internal that automatically waits for the
        job to finish, use exec_command_wait.

        :param (str) command: the command to execute
        :param (bool) combine_stderr: (default False) if True, combine stdout and
                stderr on the same buffer (i.e., stdout).
                Note: If combine_stderr is True, stderr will always be empty.
        :param (int) bufsize: same meaning of the one used by paramiko.
        
        :return: a tuple with (stdin, stdout, stderr, channel),
            where stdin, stdout and stderr behave as file-like objects,
            plus the methods provided by paramiko, and channel is a 
            paramiko.Channel object.
        """
        channel = self.sshclient.get_transport().open_session()
        channel.set_combine_stderr(combine_stderr)

        if self.getcwd() is not None:
            escaped_folder = escape_for_bash(self.getcwd())
            command_to_execute = ("cd {escaped_folder} && "
                                  "{real_command}".format(
                    escaped_folder = escaped_folder,
                    real_command = command))
        else:
            command_to_execute = command

        self.logger.debug("Command to be executed: {}".format(
                command_to_execute))

        channel.exec_command(command_to_execute)
        
        stdin = channel.makefile('wb',bufsize) 
        stdout = channel.makefile('rb',bufsize) 
        stderr = channel.makefile_stderr('rb',bufsize) 

        return stdin, stdout, stderr, channel

    def exec_command_wait(self,command,stdin=None,combine_stderr=False,
                          bufsize=-1):
        """
        Executes the specified command and waits for it to finish.
        
        :param (str) command: the command to execute
        :param (str) stdin: (optional,default=None) can be a string or a
                   file-like object.
        :param (bool) combine_stderr: (optional, default=False) see docstring of
                   self._exec_command_internal()
        :param (int) bufsize: same meaning of paramiko.

        :return: a tuple with (return_value, stdout, stderr) where stdout and stderr
            are strings.
        """
        #TODO: To see if like this it works or hangs because of buffer problems.
        
        ssh_stdin, stdout, stderr, channel = self._exec_command_internal(
            command, combine_stderr,bufsize=bufsize)
        
        if stdin is not None:
            if isinstance(stdin, basestring):
                filelike_stdin = StringIO.StringIO(stdin)
            else:
                filelike_stdin = stdin

            try:
                for l in filelike_stdin.readlines():
                    ssh_stdin.write(l)
            except AttributeError:
                raise ValueError("stdin can only be either a string of a "
                                 "file-like object!")

        # I flush and close them anyway; important to call shutdown_write
        # to avoid hangouts
        ssh_stdin.flush()
        ssh_stdin.channel.shutdown_write()

        output_text = stdout.read()
        
        # I get the return code (blocking, but in principle I waited above)
        retval = channel.recv_exit_status()

        stderr_text = stderr.read()

        return retval, output_text, stderr_text

    def gotocomputer_command(self, remotedir):
        """
        Specific gotocomputer string to connect to a given remote computer via
        ssh and directly go to the calculation folder.
        """
        
        #TODO: add also ProxyCommand and Timeout support

        further_params = []
        if 'username' in self._connect_args:
            further_params.append("-l {}".format(escape_for_bash(
                self._connect_args['username'])))
        
        if 'port' in self._connect_args:
            further_params.append("-p {}".format(self._connect_args['port']))

        if 'key_filename' in self._connect_args:
            further_params.append("-i {}".format(escape_for_bash(
                self._connect_args['key_filename'])))
        
        further_params_str = " ".join(further_params)
        connect_string = """ssh -t {machine} {further_params} "if [ -d {escaped_remotedir} ] ; then cd {escaped_remotedir} ; bash -l ; else echo '  ** The directory' ; echo '  ** {remotedir}' ; echo '  ** seems to have been deleted, I logout...' ; fi" """.format(
            further_params=further_params_str,
            machine=self._machine,
            escaped_remotedir="'{}'".format(remotedir),
            remotedir=remotedir)

        # print connect_string
        return connect_string
        
        #return """ssh -Y -t {machine} "if [ -d {escaped_remotedir} ] ; then cd {escaped_remotedir} ; bash -c "{bash_call_escaped}" ; else echo '  ** The directory' ; echo '  ** {remotedir}' ; echo '  ** seems to have been deleted, I logout...' ; fi" """.format(
        #    machine=self._machine, escaped_remotedir="'{}'".format(remotedir), remotedir=remotedir,
        #    bash_call_escaped=escape_for_bash("""bash --rcfile <(echo 'if [ -e ~/.bashrc ] ; then source ~/.bashrc ; fi ; export PS1="\[\033[01;31m\][AiiDA]\033[00m\]$PS1"')"""))
        
    def symlink(self,remotesource,remotedestination):
        """
        Create a symbolic link between the remote source and the remote 
        destination
        
        :param (str) remotesource: remote source. Can contain a pattern. 
        :param (str) remotedestination:remote destination
        """
        # paramiko gives some errors if path is starting with '.'
        s = os.path.normpath(remotesource)
        d = os.path.normpath(remotedestination)
        
        if self.has_magic(s):
            if self.has_magic(d):
                # if there are patterns in dest, I don't know which name to assign
                raise ValueError("Remotedestination cannot have patterns")
            
            # find all files matching pattern
            for this_s in self.glob(s):
                # create the name of the link: take the last part of the path
                this_d = os.path.join(remotedestination,
                                      os.path.split(this_s)[-1])
                self.sftp.symlink(this_s, this_d)
        else:
            self.sftp.symlink(s,d)
    
    def path_exists(self,path):
        """
        Check if path exists
        """
        import errno
        try:
            self.sftp.stat(path)
        except IOError, e:
            if e.errno == errno.ENOENT:
                return False
            raise
        else:
            return True
    



# -*- coding: utf-8 -*-
from aiida.parsers.plugins.quantumespresso import QEOutputParsingError
from xml.dom.minidom import parseString
from aiida.parsers.plugins.quantumespresso.basic_raw_parser_pw import (read_xml_card,
                                                                       parse_xml_child_integer, parse_xml_child_bool,
                                                                       parse_xml_child_str, parse_xml_child_float,
                                                                       parse_xml_child_attribute_str, xml_card_cell,
                                                                       xml_card_ions,
)

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


def parse_cp_traj_stanzas(num_elements, splitlines, prepend_name, rescale=1.):
    """
    num_elements: Number of lines (with three elements) between lines with two only
    elements (containing step number and time in ps).
    num_elements is 3 for cell, and the number of atoms for coordinates and positions.

    splitlines: a list of lines of the file, already split in pieces using string.split

    prepend_name: a string to be prepended to the name of keys returned
    in the return dictionary.

    rescale: the values in each stanza are multiplied by this factor, for units conversion
    """
    steps = []
    times = []
    stanzas = []
    this_stanza = []
    start_stanza = False
    linenum = -1
    try:
        for linenum, l in enumerate(splitlines):
            if len(l) == 2:
                steps.append(int(l[0]))
                times.append(float(l[1]))
                start_stanza = True
                if len(this_stanza) != 0:
                    raise ValueError("Wrong position of short line.")
            elif len(l) == 3:
                if len(this_stanza) == 0 and not start_stanza:
                    raise ValueError("Wrong position of long line.")
                start_stanza = False
                this_stanza.append([float(l[0]) * rescale, float(l[1]) * rescale, float(l[2]) * rescale])
                if len(this_stanza) == num_elements:
                    stanzas.append(this_stanza)
                    this_stanza = []
            else:
                raise ValueError("Wrong line length ({})".format(len(l)))
        if len(this_stanza) != 0:
            raise ValueError("Wrong length of last block ({} lines instead of 0)."
                             .format(len(this_stanza)))
        if len(steps) != len(stanzas):
            raise ValueError("Length mismatch between number of steps and number of defined stanzas.")
        return {
            '{}_steps'.format(prepend_name): steps,
            '{}_times'.format(prepend_name): times,
            '{}_data'.format(prepend_name): stanzas,
        }
    except Exception as e:
        e.message = "At line {}: {}".format(linenum + 1, e.message)
        raise e


def parse_cp_text_output(data, xml_data):
    """
    data must be a list of strings, one for each lines, as returned by readlines(). 
    On output, a dictionary with parsed values
    """
    # TODO: uniform readlines() and read() usage for passing input to the parser

    parsed_data = {}
    parsed_data['warnings'] = []

    for count, line in enumerate(data):

        if 'warning' in line.lower():
            parsed_data['warnings'].append(line)
        elif 'bananas' in line:
            parsed_data['warnings'].append('Bananas from the ortho.')

    for count, line in enumerate(reversed(data)):
        if 'nfi' in line and 'ekinc' in line and 'econs' in line:
            this_line = data[len(data) - count]
            try:
                parsed_data['ekinc'] = [float(this_line.split()[1])]
            except ValueError:
                pass
            try:
                parsed_data['temph'] = [float(this_line.split()[2])]
            except ValueError:
                pass
            try:
                parsed_data['tempp'] = [float(this_line.split()[3])]
            except ValueError:
                pass
            try:
                parsed_data['etot'] = [float(this_line.split()[4])]
            except ValueError:
                pass
            try:
                parsed_data['enthal'] = [float(this_line.split()[5])]
            except ValueError:
                pass
            try:
                parsed_data['econs'] = [float(this_line.split()[6])]
            except ValueError:
                pass
            try:
                parsed_data['econt'] = [float(this_line.split()[7])]
            except ValueError:
                pass
            try:
                parsed_data['vnhh'] = [float(this_line.split()[8])]
            except (ValueError, IndexError):
                pass
            try:
                parsed_data['xnhh0'] = [float(this_line.split()[9])]
            except (ValueError, IndexError):
                pass
            try:
                parsed_data['vnhp'] = [float(this_line.split()[10])]
            except (ValueError, IndexError):
                pass
            try:
                parsed_data['xnhp0'] = [float(this_line.split()[11])]
            except (ValueError, IndexError):
                pass

    return parsed_data


def parse_cp_xml_counter_output(data):
    """
    Parse xml file print_counter.xml
    data must be a single string, as returned by file.read() (notice the
    difference with parse_text_output!)
    On output, a dictionary with parsed values.
    """
    dom = parseString(data)
    parsed_data = {}
    cardname = 'LAST_SUCCESSFUL_PRINTOUT'

    card1 = [_ for _ in dom.childNodes if _.nodeName == 'PRINT_COUNTER'][0]
    card2 = [_ for _ in card1.childNodes if _.nodeName == 'LAST_SUCCESSFUL_PRINTOUT'][0]

    tagname = 'STEP'
    parsed_data[cardname.lower().replace('-', '_')] = parse_xml_child_integer(tagname, card2)

    return parsed_data


def parse_cp_raw_output(out_file, xml_file=None, xml_counter_file=None):
    parser_version = '0.1'
    parser_info = {}
    parser_info['parser_warnings'] = []
    parser_info['parser_info'] = 'AiiDA QE Parser v{}'.format(parser_version)

    # analyze the xml
    if xml_file is not None:
        try:
            with open(xml_file, 'r') as f:
                xml_lines = f.read()
        except IOError:
            raise QEOutputParsingError("Failed to open xml file: %s."
                                       .format(xml_file))
        # TODO: this function should probably be the same of pw.
        # after all, the parser was fault-tolerant
        xml_data = parse_cp_xml_output(xml_lines)
    else:
        parser_info['parser_warnings'].append('Skipping the parsing of the xml file.')
        xml_data = {}


    # analyze the counter file, which keeps info on the steps
    if xml_counter_file is not None:
        try:
            with open(xml_counter_file, 'r') as f:
                xml_counter_lines = f.read()
        except IOError:
            raise QEOutputParsingError("Failed to open xml counter file: %s."
                                       .format(xml_file))
        xml_counter_data = parse_cp_xml_counter_output(xml_counter_lines)
    else:
        xml_counter_data = {}

    # analyze the standard output
    try:
        with open(out_file, 'r') as f:
            out_lines = f.readlines()
    except IOError:
        raise QEOutputParsingError("Failed to open output file: %s." % out_file)

    # understand if the job ended smoothly
    job_successful = False
    for line in reversed(out_lines):
        if 'JOB DONE' in line:
            job_successful = True
            break

    out_data = parse_cp_text_output(out_lines, xml_data)

    for key in out_data.keys():
        if key in xml_data.keys():
            raise AssertionError('%s found in both dictionaries' % key)
        if key in xml_counter_data.keys():
            raise AssertionError('%s found in both dictionaries' % key)
            # out_data keys take precedence and overwrite xml_data keys,
            # if the same key name is shared by both (but this should not happen!)
    final_data = dict(xml_data.items() + out_data.items() + xml_counter_data.items())

    # TODO: parse the trajectory and save them in a reasonable format

    return final_data, job_successful


# TODO: the xml has a lot in common with pw, maybe I should avoid duplication of code
# or maybe should I wait for the new version of data-file.xml ?
def parse_cp_xml_output(data):
    """
    Parse xml data
    data must be a single string, as returned by file.read() (notice the
    difference with parse_text_output!)
    On output, a dictionary with parsed values.
    Democratically, we have decided to use picoseconds as units of time, eV for energies, Angstrom for lengths.
    """
    import copy

    dom = parseString(data)

    parsed_data = {}

    # CARD STATUS

    cardname = 'STATUS'
    target_tags = read_xml_card(dom, cardname)

    tagname = 'STEP'
    attrname = 'ITERATION'
    parsed_data[(tagname + '_' + attrname).lower()] = int(parse_xml_child_attribute_str(tagname, attrname, target_tags))

    tagname = 'TIME'
    attrname = 'UNITS'
    value = parse_xml_child_float(tagname, target_tags)
    units = parse_xml_child_attribute_str(tagname, attrname, target_tags)
    if units not in ['pico-seconds']:
        raise QEOutputParsingError("Units {} are not supported by parser".format(units))
    parsed_data[tagname.lower()] = value

    tagname = 'TITLE'
    parsed_data[tagname.lower()] = parse_xml_child_str(tagname, target_tags)

    # CARD CELL
    parsed_data, lattice_vectors, volume = copy.deepcopy(xml_card_cell(parsed_data, dom))

    # CARD IONS
    parsed_data = copy.deepcopy(xml_card_ions(parsed_data, dom, lattice_vectors, volume))

    # CARD TIMESTEPS

    cardname = 'TIMESTEPS'
    target_tags = read_xml_card(dom, cardname)

    for tagname in ['STEP0', 'STEPM']:
        try:
            tag = target_tags.getElementsByTagName(tagname)[0]

            try:
                second_tagname = 'ACCUMULATORS'
                second_tag = tag.getElementsByTagName(second_tagname)[0]
                data = second_tag.childNodes[0].data.rstrip().split()  # list of floats
                parsed_data[second_tagname.replace('-', '_').lower()] = [float(i) for i in data]
            except:
                pass

            second_tagname = 'IONS_POSITIONS'
            second_tag = tag.getElementsByTagName(second_tagname)[0]
            third_tagname = 'stau'
            third_tag = second_tag.getElementsByTagName(third_tagname)[0]
            list_data = third_tag.childNodes[0].data.rstrip().split()
            list_data = [float(i) for i in list_data]
            # convert to matrix
            val = []
            mat = []
            for i, data in enumerate(list_data):
                val.append(data)
                if (i + 1) % 3 == 0:
                    mat.append(val)
                    val = []
            parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            third_tagname = 'svel'
            third_tag = second_tag.getElementsByTagName(third_tagname)[0]
            list_data = third_tag.childNodes[0].data.rstrip().split()
            list_data = [float(i) for i in list_data]
            # convert to matrix
            val = []
            mat = []
            for i, data in enumerate(list_data):
                val.append(data)
                if (i + 1) % 3 == 0:
                    mat.append(val)
                    val = []
            parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            try:
                third_tagname = 'taui'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]
                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass

            try:
                third_tagname = 'cdmi'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = [float(i) for i in
                                                                                                 list_data]
            except:
                pass

            try:
                third_tagname = 'force'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]
                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass

            second_tagname = 'IONS_NOSE'
            second_tag = tag.getElementsByTagName(second_tagname)[0]
            third_tagname = 'nhpcl'
            third_tag = second_tag.getElementsByTagName(third_tagname)[0]
            parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = float(
                third_tag.childNodes[0].data)
            third_tagname = 'nhpdim'
            third_tag = second_tag.getElementsByTagName(third_tagname)[0]
            parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = float(
                third_tag.childNodes[0].data)
            third_tagname = 'xnhp'
            third_tag = second_tag.getElementsByTagName(third_tagname)[0]
            parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = float(
                third_tag.childNodes[0].data)
            try:
                third_tagname = 'vnhp'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = float(
                    third_tag.childNodes[0].data)
            except:
                pass

            try:
                second_tagname = 'ekincm'
                second_tag = tag.getElementsByTagName(second_tagname)[0]
                parsed_data[second_tagname.replace('-', '_').lower()] = float(second_tag.childNodes[0].data)
            except:
                pass

            second_tagname = 'ELECTRONS_NOSE'
            second_tag = tag.getElementsByTagName(second_tagname)[0]
            try:
                third_tagname = 'xnhe'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = float(
                    third_tag.childNodes[0].data)
            except:
                pass
            try:
                third_tagname = 'vnhe'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = float(
                    third_tag.childNodes[0].data)
            except:
                pass

            second_tagname = 'CELL_PARAMETERS'
            second_tag = tag.getElementsByTagName(second_tagname)[0]
            try:
                third_tagname = 'ht'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]
                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass
            try:
                third_tagname = 'htvel'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]
                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass
            try:
                third_tagname = 'gvel'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]
                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass

            second_tagname = 'CELL_NOSE'
            second_tag = tag.getElementsByTagName(second_tagname)[0]
            try:
                third_tagname = 'xnhh'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]

                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass
            try:
                third_tagname = 'vnhh'
                third_tag = second_tag.getElementsByTagName(third_tagname)[0]
                list_data = third_tag.childNodes[0].data.rstrip().split()
                list_data = [float(i) for i in list_data]
                # convert to matrix
                val = []
                mat = []
                for i, data in enumerate(list_data):
                    val.append(data)
                    if (i + 1) % 3 == 0:
                        mat.append(val)
                        val = []
                parsed_data[(second_tagname + '_' + third_tagname).replace('-', '_').lower()] = mat
            except:
                pass
        except:
            raise QEOutputParsingError('Error parsing CARD {}'.format(cardname))

    # CARD BAND_STRUCTURE_INFO

    cardname = 'BAND_STRUCTURE_INFO'
    target_tags = read_xml_card(dom, cardname)

    tagname = 'NUMBER_OF_ATOMIC_WFC'
    parsed_data[tagname.lower().replace('-', '_')] = parse_xml_child_integer(tagname, target_tags)

    tagname = 'NUMBER_OF_ELECTRONS'
    parsed_data[tagname.lower().replace('-', '_')] = int(parse_xml_child_float(tagname, target_tags))

    tagname = 'NUMBER_OF_BANDS'
    parsed_data[tagname.lower().replace('-', '_')] = parse_xml_child_integer(tagname, target_tags)

    tagname = 'NUMBER_OF_SPIN_COMPONENTS'
    parsed_data[tagname.lower().replace('-', '_')] = parse_xml_child_integer(tagname, target_tags)

    return parsed_data



# -*- coding: utf-8 -*-

import unittest
from aiida.common.extendeddicts import DefaultsDict


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

class TestDefaultsDict(unittest.TestCase):
    def setUp(self):
        self.defaults_dict = DefaultsDict(
            valid_keys=['foo', 'bar'],
            defaults={'bar': 'bar_default'})

    def test_setattr(self):
        # Test setting and getting a value
        self.defaults_dict.foo = 'hello'
        self.assertEqual(self.defaults_dict.foo, 'hello')

        # Test setting an invalid attribute
        with self.assertRaises(AttributeError):
            self.defaults_dict.non_existent = None

    def test_defaults(self):
        self.assertEquals(self.defaults_dict.defaults, {'bar': 'bar_default'})

    def test_getattr(self):
        # Test getting an unset value
        with self.assertRaises(AttributeError):
            self.defaults_dict.unset

    def test_delattr(self):
        self.defaults_dict.foo = 'hello'
        del self.defaults_dict.foo
        # Now try deleting it again
        with self.assertRaises(AttributeError):
            del self.defaults_dict.foo

        # Try deleting one that never existed
        with self.assertRaises(AttributeError):
            del self.defaults_dict.foo

    def test_delitem(self):
        self.defaults_dict['foo'] = 'test'
        del self.defaults_dict['foo']
        # Try deleting again
        with self.assertRaises(KeyError):
            del self.defaults_dict['foo']

        # Try deleting on that never existed
        with self.assertRaises(KeyError):
            del self.defaults_dict['non_existent']

    def test_invalid_default(self):
        with self.assertRaises(AssertionError):
            DefaultsDict([], defaults={'foo': 'bar'})

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
from abc import abstractmethod, abstractproperty, ABCMeta

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class AbstractComment(object):
    __metaclass__ = ABCMeta

    _logger = logging.getLogger(__name__)

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractproperty
    def pk(self):
        pass

    @abstractproperty
    def id(self):
        pass

    @abstractproperty
    def to_be_stored(self):
        pass

    @abstractproperty
    def uuid(self):
        pass

    @abstractmethod
    def get_ctime(self):
        pass

    @abstractmethod
    def set_ctime(self, val):
        pass

    @abstractmethod
    def get_mtime(self):
        pass

    @abstractmethod
    def set_mtime(self, val):
        pass

    @abstractmethod
    def get_user(self):
        pass

    @abstractmethod
    def set_user(self, val):
        pass

    @abstractmethod
    def get_content(self):
        pass

    @abstractmethod
    def set_content(self, val):
        pass

    @abstractmethod
    def delete(self):
        pass



# -*- coding: utf-8 -*-

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

from aiida.tools.dbexporters.tcod_plugins import BaseTcodtranslator

class CpTcodtranslator(BaseTcodtranslator):
    """
    Quantum ESPRESSO's CP-specific input and output parameter translator
    to TCOD CIF dictionary tags.
    """
    _plugin_type_string = "quantumespresso.cp.CpCalculation"

    @classmethod
    def get_software_package(cls,calc,**kwargs):
        """
        Returns the package or program name that was used to produce
        the structure. Only package or program name should be used,
        e.g. 'VASP', 'psi3', 'Abinit', etc.
        """
        return 'Quantum ESPRESSO'

    @classmethod
    def get_number_of_electrons(cls,calc,**kwargs):
        """
        Returns the number of electrons.
        """
        parameters = calc.out.output_parameters
        if 'number_of_electrons' not in parameters.attrs():
            return None
        return parameters.get_attr('number_of_electrons')

    @classmethod
    def get_computation_wallclock_time(cls,calc,**kwargs):
        """
        Returns the computation wallclock time in seconds.
        """
        parameters = calc.out.output_parameters
        if 'wall_time_seconds' not in parameters.attrs():
            return None
        return parameters.get_attr('wall_time_seconds')



# -*- coding: utf-8 -*-
"""
This file contains tests for AiiDA.

This module checks that the module is properly loaded (i.e., if the
settings.aiida_test_list variable is set) and dynamically runs only the tests
specified in this list. See also the 'verdi test' function.
"""
import sys
import inspect
import importlib

from django.utils import unittest

from aiida.common.exceptions import InternalError
from aiida.backends.djsite.settings import settings_profile
from aiida.backends.djsite.db.testbase import db_test_list

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

tests_to_run = settings_profile.__dict__.get('aiida_test_list', None)
if tests_to_run is None:
    # This is intended as a safety mechanism. This should help to ensure that
    # tests are  run only after setting the flag in the djsite.settings.settings
    # module, and more importantly only after changing the DB and REPO to be
    # test ones.
    raise InternalError("aiida_test_list is not set, but you are trying to run the tests...")

actually_run_tests = []
num_tests_expected = 0
for test in set(tests_to_run):
    try:
        modulenames = db_test_list[test]
    except KeyError:
        print >> sys.stderr, "Unknown DB test {}... skipping".format(test)
        continue

    for modulename in modulenames:
        module = importlib.import_module(modulename)
        for objname, obj in module.__dict__.iteritems():
            if inspect.isclass(obj) and issubclass(obj, unittest.TestCase):
                # See if the object has test methods.
                # Add only if there is at least one test_ method
                testmethods = [
                    m for m in inspect.getmembers(obj, predicate=inspect.ismethod)
                    if m[0].startswith("test_")]
                if testmethods:  # at least a method starting with test_
                    if objname in locals():
                        raise InternalError(
                            "Test class {} defined more than once".format(objname))
                    locals()[objname] = obj
                    num_tests_expected += len(testmethods)
                    # for debug
                    #print "{} ==> {} ({})".format(modulename, objname, len(testmethods))
        actually_run_tests.append(test)

obj = None  # To avoid double runnings of the last test

print >> sys.stderr, "DB tests that will be run: {} (expecting {} tests)".format(
    ",".join(actually_run_tests), num_tests_expected)




# -*- coding: utf-8 -*-
"""
This allows to setup and configure a code from command line.

TODO: think if we want to allow to change path and prepend/append text.
"""
import sys

from aiida.cmdline.baseclass import VerdiCommandWithSubcommands

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


def cmdline_fill(attributes, store, print_header=True):
    import inspect
    import readline
    from aiida.common.exceptions import ValidationError

    if print_header:
        print "At any prompt, type ? to get some help."
        print "---------------------------------------"

    for internal_name, name, desc, multiline in (
            attributes):

        getter_name = '_get_{}_string'.format(internal_name)
        try:
            getter = dict(inspect.getmembers(
                store))[getter_name]
        except KeyError:
            print >> sys.stderr, ("Internal error! "
                                  "No {} getter defined in Computer".format(getter_name))
            sys.exit(1)
        previous_value = getter()

        setter_name = '_set_{}_string'.format(internal_name)
        try:
            setter = dict(inspect.getmembers(
                store))[setter_name]
        except KeyError:
            print >> sys.stderr, ("Internal error! "
                                  "No {} setter defined in Computer".format(setter_name))
            sys.exit(1)

        valid_input = False
        while not valid_input:
            if multiline:
                newlines = []
                print "=> {}: ".format(name)
                print "   # This is a multiline input, press CTRL+D on a"
                print "   # empty line when you finish"

                try:
                    for l in previous_value.splitlines():
                        while True:
                            readline.set_startup_hook(lambda:
                                                      readline.insert_text(l))
                            input_txt = raw_input()
                            if input_txt.strip() == '?':
                                print ["  > {}".format(descl) for descl
                                       in "HELP: {}".format(desc).split('\n')]
                                continue
                            else:
                                newlines.append(input_txt)
                                break

                    # Reset the hook (no default text printed)
                    readline.set_startup_hook()

                    print "   # ------------------------------------------"
                    print "   # End of old input. You can keep adding     "
                    print "   # lines, or press CTRL+D to store this value"
                    print "   # ------------------------------------------"

                    while True:
                        input_txt = raw_input()
                        if input_txt.strip() == '?':
                            print "\n".join(["  > {}".format(descl) for descl
                                             in "HELP: {}".format(desc).split('\n')])
                            continue
                        else:
                            newlines.append(input_txt)
                except EOFError:
                    # Ctrl+D pressed: end of input.
                    pass

                input_txt = "\n".join(newlines)

            else:  # No multiline
                readline.set_startup_hook(lambda: readline.insert_text(
                    previous_value))
                input_txt = raw_input("=> {}: ".format(name))
                if input_txt.strip() == '?':
                    print "HELP:", desc
                    continue

            try:
                setter(input_txt)
                valid_input = True
            except ValidationError as e:
                print >> sys.stderr, "Invalid input: {}".format(e.message)
                print >> sys.stderr, "Enter '?' for help".format(e.message)


class CodeInputValidationClass(object):
    """
    A class with information for the validation of input text of Codes
    """
    # It is a list of tuples. Each tuple has three elements:
    # 1. an internal name (used to find the
    # _set_internalname_string, and get_internalname_string methods)
    # 2. a short human-readable name
    # 3. A long human-readable description
    # 4. True if it is a multi-line input, False otherwise
    # IMPORTANT!
    # for each entry, remember to define the
    # _set_internalname_string and get_internalname_string methods.
    # Moreover, the _set_internalname_string method should also immediately
    # validate the value.
    _conf_attributes_relabel = [
        ("label",
         "Label",
         "A label to refer to this code",
         False,
        ),
        ("description",
         "Description",
         "A human-readable description of this code",
         False,
        ),
    ]
    _conf_attributes_start = [
        ("label",
         "Label",
         "A label to refer to this code",
         False,
        ),
        ("description",
         "Description",
         "A human-readable description of this code",
         False,
        ),
        ("is_local",
         "Local",
         "True or False; if True, then you have to provide a folder with "
         "files that will be stored in AiiDA and copied to the remote "
         "computers for every calculation submission. If True, the code "
         "is just a link to a remote computer and an absolute path there",
         False,
        ),
        ("input_plugin",
         "Default input plugin",
         "A string of the default input plugin to be used with this code "
         "that is recognized by the CalculationFactory. Use the "
         "'verdi calculation plugins' command to get the list of existing"
         "plugins",
         False,
        ),
    ]
    _conf_attributes_local = [
        ("folder_with_code",
         "Folder with the code",
         "The folder on your local computer in which there are the files to be "
         "stored in the AiiDA repository and then copied over for every "
         "submitted calculation",
         False,
        ),
        ("local_rel_path",
         "Relative path of the executable",
         "The relative path of the executable file inside the folder entered "
         "in the previous step",
         False,
        ),
    ]
    _conf_attributes_remote = [
        ("computer",
         "Remote computer name",
         "The computer name as on which the code resides, as stored in the "
         "AiiDA database",
         False,
        ),
        ("remote_abs_path",
         "Remote absolute path",
         "The (full) absolute path on the remote machine",
         False,
        ),
    ]
    _conf_attributes_end = [
        ("prepend_text",
         "Text to prepend to each command execution\n"
         "FOR INSTANCE, MODULES TO BE LOADED FOR THIS CODE",
         "This is a multiline string, whose content will be prepended inside\n"
         "the submission script before the real execution of the job. It is\n"
         "your responsibility to write proper bash code!",
         True,
        ),
        ("append_text",
         "Text to append to each command execution",
         "This is a multiline string, whose content will be appended inside\n"
         "the submission script after the real execution of the job. It is\n"
         "your responsibility to write proper bash code!",
         True,
        ),
    ]

    label = ""

    def _get_label_string(self):
        return self.label

    def _set_label_string(self, string):
        """
        Set the label starting from a string.
        """
        self._label_validator(string)
        self.label = string

    def _label_validator(self, label):
        """
        Validates the label.
        """
        from aiida.common.exceptions import ValidationError

        if not label.strip():
            raise ValidationError("No label specified")

    description = ""

    def _get_description_string(self):
        return self.description

    def _set_description_string(self, string):
        """
        Set the description starting from a string.
        """
        self._description_validator(string)
        self.description = string

    def _description_validator(self, folder_with_code):
        """
        Validates the folder_with_code.
        """
        pass

    folder_with_code = ""

    def _get_folder_with_code_string(self):
        return self.folder_with_code

    def _set_folder_with_code_string(self, string):
        """
        Set the folder_with_code starting from a string.
        """
        self._folder_with_code_validator(string)
        self.folder_with_code = string

    def _folder_with_code_validator(self, folder_with_code):
        """
        Validates the folder_with_code.
        """
        import os.path
        from aiida.common.exceptions import ValidationError

        if not os.path.isdir(folder_with_code):
            raise ValidationError("'{}' is not a valid directory".format(
                folder_with_code))

    local_rel_path = ""

    def _get_local_rel_path_string(self):
        return self.local_rel_path

    def _set_local_rel_path_string(self, string):
        """
        Set the local_rel_path starting from a string.
        """
        self._local_rel_path_validator(string)
        self.local_rel_path = string

    def _local_rel_path_validator(self, local_rel_path):
        """
        Validates the local_rel_path.
        """
        import os.path
        from aiida.common.exceptions import ValidationError

        if not os.path.isfile(os.path.join(self.folder_with_code,
                                           local_rel_path)):
            raise ValidationError("'{}' is not a valid file within '{}'".format(
                local_rel_path, self.folder_with_code))

    computer = None

    def _get_computer_string(self):
        if self.computer is None:
            return ""
        else:
            return self.computer.name

    def _set_computer_string(self, string):
        """
        Set the computer starting from a string.
        """
        from aiida.common.exceptions import ValidationError, NotExistent
        from aiida.orm import Computer as AiidaOrmComputer

        try:
            computer = AiidaOrmComputer.get(string)
        except NotExistent:
            raise ValidationError("Computer with name '{}' not found in "
                                  "DB".format(string))

        self._computer_validator(computer)
        self.computer = computer

    def _computer_validator(self, computer):
        """
        Validates the computer.
        """
        from aiida.common.exceptions import ValidationError
        from aiida.orm import Computer as AiidaOrmComputer

        if not isinstance(computer, AiidaOrmComputer):
            raise ValidationError("The computer is not a valid Computer instance")

    remote_abs_path = ""

    def _get_remote_abs_path_string(self):
        return self.remote_abs_path

    def _set_remote_abs_path_string(self, string):
        """
        Set the remote_abs_path starting from a string.
        """
        self._remote_abs_path_validator(string)
        self.remote_abs_path = string

    def _remote_abs_path_validator(self, remote_abs_path):
        """
        Validates the remote_abs_path.
        """
        from aiida.common.exceptions import ValidationError
        import os.path

        if not os.path.isabs(remote_abs_path):
            raise ValidationError("This is not a valid absolute path")
        if not os.path.split(remote_abs_path)[1]:
            raise ValidationError("This is a folder, not an executable")

    is_local = False

    def _get_is_local_string(self):
        return "True" if self.is_local else "False"

    def _set_is_local_string(self, string):
        """
        Set the is_local starting from a string.
        """
        from aiida.common.exceptions import ValidationError

        upper_string = string.upper()
        if upper_string in ['YES', 'Y', 'T', 'TRUE']:
            is_local = True
        elif upper_string in ['NO', 'N', 'F', 'FALSE']:
            is_local = False
        else:
            raise ValidationError("Invalid value '{}' for the is_local variable, must "
                                  "be a boolean".format(string))

        self._is_local_validator(is_local)
        self.is_local = is_local

    def _is_local_validator(self, is_local):
        """
        Validates the is_local.
        """
        from aiida.common.exceptions import ValidationError

        if not isinstance(is_local, bool):
            raise ValidationError("Invalid value '{}' for the is_local variable, must "
                                  "be a boolean".format(str(is_local)))

    input_plugin = None

    def _get_input_plugin_string(self):
        """
        Return the input plugin string
        """
        return self.input_plugin

    def _set_input_plugin_string(self, string):
        """
        Set the input_plugin starting from a string.
        """
        input_plugin = string.strip()

        if input_plugin.lower == "none":
            input_plugin = None

        self._input_plugin_validator(input_plugin)
        self.input_plugin = input_plugin

    def _input_plugin_validator(self, input_plugin):
        """
        Validates the input_plugin, checking it is in the list of existing
        plugins.
        """
        from aiida.common.exceptions import ValidationError
        from aiida.orm import JobCalculation
        from aiida.common.pluginloader import existing_plugins

        if input_plugin is None:
            return

        if input_plugin not in existing_plugins(JobCalculation,
                                                'aiida.orm.calculation.job',
                                                suffix='Calculation'):
            raise ValidationError("Invalid value '{}' for the input_plugin "
                                  "variable, it is not among the existing plugins".format(
                str(input_plugin)))

    prepend_text = ""

    def _get_prepend_text_string(self):
        return self.prepend_text

    def _set_prepend_text_string(self, string):
        """
        Set the prepend_text starting from a string.
        """
        self._prepend_text_validator(string)
        self.prepend_text = string

    def _prepend_text_validator(self, prepend_text):
        """
        Validates the prepend_text.
        """
        pass

    append_text = ""

    def _get_append_text_string(self):
        return self.append_text

    def _set_append_text_string(self, string):
        """
        Set the append_text starting from a string.
        """
        self._append_text_validator(string)
        self.append_text = string

    def _append_text_validator(self, append_text):
        """
        Validates the append_text.
        """
        pass

    def create_code(self):
        """
        Create a code with the information contained in this class,
        BUT DOES NOT STORE IT.
        """
        import os.path
        from aiida.orm import Code as AiidaOrmCode

        if self.is_local:
            file_list = [os.path.realpath(os.path.join(self.folder_with_code, f))
                         for f in os.listdir(self.folder_with_code)]
            code = AiidaOrmCode(local_executable=self.local_rel_path,
                                files=file_list)
        else:
            code = AiidaOrmCode(remote_computer_exec=(self.computer,
                                                      self.remote_abs_path))

        code.label = self.label
        code.description = self.description
        code.set_input_plugin_name(self.input_plugin)
        code.set_prepend_text(self.prepend_text)
        code.set_append_text(self.append_text)

        return code

    #     def load_from_code(self, code):
    #         from aiida.orm import Code as AiidaOrmCode
    #
    #         if not isinstance(code, AiidaOrmCode):
    #             raise ValueError("code is not a valid Code instance")
    #
    #         self.label = code.label
    #         self.description = code.description
    #         # Add here also the input_plugin stuff
    #         self.is_local = code.is_local()
    #         if self.is_local:
    #             raise NotImplementedError
    #         else:
    #             self.computer = code.get_remote_computer()
    #             self.remote_abs_path = code.get_remote_exec_path()
    #         self.prepend_text = code.get_prepend_text()
    #         self.append_text = code.get_append_text()

    def set_and_validate_from_code(self, kwargs):
        """
        This method is used by the Code Orm, for the utility to setup a new code
        from the verdi shell
        """
        from aiida.common.exceptions import ValidationError

        # convert to string so I can use all the functionalities of the command line
        kwargs = { k:str(v) for k,v in kwargs.iteritems() }

        start_var = [ _[0] for _ in self._conf_attributes_start ]
        local_var = [ _[0] for _ in self._conf_attributes_local ]
        remote_var = [ _[0] for _ in self._conf_attributes_remote ]
        end_var = [ _[0] for _ in self._conf_attributes_end ]

        def internal_launch(self, x, kwargs):
            default_values = { k:getattr(self,'_get_{}_string'.format(k))() for k in x }
            setup_keys = [ [k,kwargs.pop(k,default_values[k]) ] for k in x ]
#            for k,v in setup_keys:
#                setattr(self,k,v)
            [ getattr(self,'_set_{}_string'.format(k))(v) for k,v in setup_keys ]

            return kwargs

        kwargs = internal_launch(self, start_var, kwargs)

        if self.is_local:
            kwargs = internal_launch(self, local_var, kwargs)
        else:
            print 'called remote', remote_var
            kwargs = internal_launch(self, remote_var, kwargs)

        kwargs = internal_launch(self, end_var, kwargs)

        print kwargs

        if kwargs:
            raise ValidationError("Some parameters were not "
                                  "recognized: {}".format(kwargs))
        return self.create_code()

    def ask(self):
        cmdline_fill(self._conf_attributes_start,
                     store=self)
        if self.is_local:
            cmdline_fill(self._conf_attributes_local,
                         store=self, print_header=False)
        else:
            cmdline_fill(self._conf_attributes_remote,
                         store=self, print_header=False)
        cmdline_fill(self._conf_attributes_end,
                     store=self, print_header=False)


class Code(VerdiCommandWithSubcommands):
    """
    Setup and manage codes to be used

    This command allows to list, add, modify, configure codes.
    """

    def __init__(self):
        from aiida.backends.utils import load_dbenv, is_dbenv_loaded
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()

        self.valid_subcommands = {
            'list': (self.code_list, self.complete_none),
            'show': (self.code_show, self.complete_code_names_and_pks),
            'setup': (self.code_setup, self.complete_code_pks),
            'rename': (self.code_rename, self.complete_none),
            'update': (self.code_update, self.complete_code_pks),
            'delete': (self.code_delete, self.complete_code_pks),
            'hide': (self.code_hide, self.complete_code_pks),
            'reveal': (self.code_reveal, self.complete_code_pks),
        }

    def complete_code_names(self, subargs_idx, subargs):
        code_names = [c[1] for c in self.get_code_data()]
        return "\n".join(code_names)

    def complete_code_pks(self, subargs_idx, subargs):
        code_pks = [str(c[0]) for c in self.get_code_data()]
        return "\n".join(code_pks)

    def complete_code_names_and_pks(self, subargs_idx, subargs):
        return "\n".join([self.complete_code_names(subargs_idx, subargs),
                          self.complete_code_pks(subargs_idx, subargs)])

    def code_hide(self, *args):
        """
        Hide one or more codes from the verdi show command
        """
        import argparse
        from aiida.orm.code import Code as OrmCode
        from aiida.orm.utils import load_node

        parser = argparse.ArgumentParser(prog=self.get_full_command_name(),
                                         description='Hide codes from the verdi show command.')
        # The default states are those that are shown if no option is given
        parser.add_argument('pks', type=int, nargs='+',
                            help="The pk of the codes to hide",
        )
        parsed_args = parser.parse_args(args)
        from aiida.orm.code import Code

        for pk in parsed_args.pks:
            code = load_node(pk, parent_class=OrmCode)
            code._hide()

    def code_reveal(self, *args):
        """
        Reveal (if it was hidden before) one or more codes from the verdi show command
        """
        import argparse
        from aiida.orm.code import Code as OrmCode
        from aiida.orm.utils import load_node

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Reveal codes (if they were hidden before) from the verdi show command.')
        # The default states are those that are shown if no option is given
        parser.add_argument('pks', type=int, nargs='+',
                            help="The pk of the codes to reveal",
        )
        parsed_args = parser.parse_args(args)
        from aiida.orm.code import Code

        for pk in parsed_args.pks:
            code = load_node(pk, parent_class=OrmCode)
            code._reveal()

    def code_list(self, *args):
        """
        List available codes
        """
        import argparse

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='List the codes in the database.')
        # The default states are those that are shown if no option is given
        parser.add_argument('-c', '--computer',
                            help="Filter only codes on a given computer",
        )
        parser.add_argument('-p', '--plugin',
                            help="Filter only calculation with a given plugin",
        )
        parser.add_argument('-A', '--all-users', dest='all_users',
                            action='store_true',
                            help="Show codes of all users",
        )
        parser.add_argument('-o', '--show-owner', dest='show_owner',
                            action='store_true',
                            help="Show also the owner of the code",
        )
        parser.add_argument('-a', '--all-codes',
                            action='store_true',
                            help="Show also hidden codes",
        )
        parser.set_defaults(all_users=False, hidden=False)
        parsed_args = parser.parse_args(args)
        computer_filter = parsed_args.computer
        plugin_filter = parsed_args.plugin
        all_users = parsed_args.all_users
        show_owner = parsed_args.show_owner
        reveal_filter = parsed_args.all_codes

        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.code import Code
        from aiida.orm.computer import Computer
        from aiida.orm.user import User
        from aiida.backends.utils import get_automatic_user

        qb_user_filters = dict()
        if not all_users:
            user = User(dbuser=get_automatic_user())
            qb_user_filters['email'] = user.email

        qb_computer_filters = dict()
        if computer_filter is not None:
            qb_computer_filters['name'] = computer_filter

        qb_code_filters = dict()
        if plugin_filter is not None:
            qb_code_filters['attributes.input_plugin'] = plugin_filter

        if not reveal_filter:
            qb_code_filters['attributes.hidden'] = {"~==": True}

        qb = QueryBuilder()
        qb.append(Code, tag="code",
                  filters=qb_code_filters,
                  project=["id", "label"])
        qb.append(Computer, computer_of="code",
                  project=["name"],
                  filters=qb_computer_filters)
        qb.append(User, creator_of="code",
                  project=["email"],
                  filters=qb_user_filters)

        print "# List of configured codes:"
        print "# (use 'verdi code show CODEID' to see the details)"
        if qb.count > 0:
            for pk, label, computername, useremail in qb.iterall():
                if show_owner:
                    owner_string = " ({})".format(useremail)
                else:
                    owner_string = ""
                if computername is None:
                    computernamestring = ""
                else:
                    computernamestring = "@{}".format(computername)
                print "* pk {} - {}{}{}".format(
                    pk, label, computernamestring, owner_string)
        else:
            print "# No codes found matching the specified criteria."

    def get_code_data(self, django_filter=None):
        """
        Retrieve the list of codes in the DB.
        Return a tuple with (pk, label, computername, owneremail).

        :param django_filter: a django query object (e.g. obtained
          with Q()) to filter the results on the AiidaOrmCode class.
        """
        from aiida.orm import Code as AiidaOrmCode
        from aiida.orm.querybuilder import QueryBuilder

        qb = QueryBuilder()
        qb.append(AiidaOrmCode, project=['id', 'label'])
        qb.append(type='computer', computer_of=AiidaOrmCode, project=['name'])
        qb.append(type='user', creator_of=AiidaOrmCode, project=['email'])

        return sorted(qb.all())

    def get_code(self, code_id):
        """
        Get a Computer object with given identifier, that can either be
        the numeric ID (pk), or the label (if unique).

        .. note:: If an string that can be converted to an integer is given,
            the numeric ID is verified first (therefore, is a code A with a
            label equal to the ID of another code B is present, code A cannot
            be referenced by label).
        """
        from aiida.common.exceptions import NotExistent, MultipleObjectsError

        from aiida.orm import Code as AiidaOrmCode

        try:
            return AiidaOrmCode.get_from_string(code_id)
        except (NotExistent, MultipleObjectsError) as e:
            print >> sys.stderr, e.message
            sys.exit(1)

    def code_show(self, *args):
        """
        Show information on a given code
        """
        if len(args) != 1:
            print >> sys.stderr, ("after 'code show' there should be one "
                                  "argument only, being the code id.")
            sys.exit(1)

        code = self.get_code(args[0])
        print code.full_text_info

    def code_setup(self, *args):
        from aiida.common.exceptions import ValidationError

        if len(args) != 0:
            print >> sys.stderr, ("after 'code setup' there cannot be any "
                                  "argument")
            sys.exit(1)

        set_params = CodeInputValidationClass()

        set_params.ask()

        code = set_params.create_code()

        # Enforcing the code to be not hidden.
        code._reveal()

        try:
            code.store()
        except ValidationError as e:
            print "Unable to store the computer: {}. Exiting...".format(e.message)
            sys.exit(1)

        print "Code '{}' successfully stored in DB.".format(code.label)
        print "pk: {}, uuid: {}".format(code.pk, code.uuid)

    def code_rename(self, *args):
        import argparse
        from aiida.common.exceptions import NotExistent

        from aiida.orm.code import Code

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Rename a code (change its label).')
        # The default states are those that are shown if no option is given
        parser.add_argument('old_name', help="The old name of the code")
        parser.add_argument('new_name', help="The new name of the code")

        parsed_args = parser.parse_args(args)

        new_name = parsed_args.new_name
        old_name = parsed_args.old_name

        try:
            code = Code.get_from_string(old_name)
        except NotExistent:
            print "ERROR! A code with name {} could not be found".format(old_name)
            sys.exit(1)

        suffix = '@{}'.format(code.get_computer().name)
        if new_name.endswith(suffix):
            new_name = new_name[:-len(suffix)]

        if '@' in new_name:
            print >> sys.stderr, "ERROR! Do not put '@' symbols in the code name"
            sys.exit(1)

        retrieved_old_name = '{}@{}'.format(code.label, code.get_computer().name)
        # CHANGE HERE
        code.label = new_name
        retrieved_new_name = '{}@{}'.format(code.label, code.get_computer().name)

        print "Renamed code with ID={} from '{}' to '{}'".format(
            code.pk, retrieved_old_name, retrieved_new_name)

    def code_update(self, *args):
        import datetime
        from aiida.backends.utils import get_automatic_user

        if len(args) != 1:
            print >> sys.stderr, ("after 'code update' there should be one "
                                  "argument only, being the code id.")
            sys.exit(1)

        code = self.get_code(args[0])

        if code.has_children:
            print "***********************************"
            print "|                                 |"
            print "|            WARNING!             |"
            print "| Consider to create another code |"
            print "| You risk of losing the history  |"
            print "|                                 |"
            print "***********************************"

        # load existing stuff
        set_params = CodeInputValidationClass()
        set_params.label = code.label
        set_params.description = code.description
        set_params.input_plugin = code.get_input_plugin_name()

        was_local_before = code.is_local()
        set_params.is_local = code.is_local()

        if code.is_local():
            set_params.local_rel_path = code.get_local_executable()
            # I don't have saved the folder with code, so I will just have the list of files
            # file_list = [ code._get_folder_pathsubfolder.get_abs_path(i)
            #    for i in code.get_folder_list() ]
        else:
            set_params.computer = code.get_computer()
            set_params.remote_abs_path = code.get_remote_exec_path()

        set_params.prepend_text = code.get_prepend_text()
        set_params.append_text = code.get_append_text()

        # ask for the new values
        set_params.ask()

        # prepare a comment containing the previous version of the code
        now = datetime.datetime.now()
        new_comment = []
        new_comment.append("Code modified on {}".format(now))
        new_comment.append("Old configuration was:")
        new_comment.append("label: {}".format(code.label))
        new_comment.append("description: {}".format(code.description))
        new_comment.append("input_plugin_name: {}".format(code.get_input_plugin_name()))
        new_comment.append("is_local: {}".format(code.is_local()))
        if was_local_before:
            new_comment.append("local_executable: {}".format(code.get_local_executable()))
        else:
            new_comment.append("computer: {}".format(code.get_computer()))
            new_comment.append("remote_exec_path: {}".format(code.get_remote_exec_path()))
        new_comment.append("prepend_text: {}".format(code.get_prepend_text()))
        new_comment.append("append_text: {}".format(code.get_append_text()))
        comment = "\n".join(new_comment)

        if set_params.is_local:
            print "WARNING: => Folder with the code, and"
            print "         => Relative path of the executable, "
            print "         will be ignored! It is not possible to replace "
            print "         the scripts, you have to create a new code for that."
        else:
            if was_local_before:
                # some old files will be left in the repository, and I cannot delete them
                print >> sys.stderr, ("It is not possible to change a "
                                      "code from local to remote.\n"
                                      "Modification cancelled.")
                sys.exit(1)
            print "WARNING: => computer"
            print "         will be ignored! It is not possible to replace it"
            print "         you have to create a new code for that."

        code.label = set_params.label
        code.description = set_params.description
        code.set_input_plugin_name(set_params.input_plugin)
        code.set_prepend_text(set_params.prepend_text)
        code.set_append_text(set_params.append_text)

        if not was_local_before:
            if set_params.remote_abs_path != code.get_remote_exec_path():
                print "Are you sure about changing the path of the code?"
                print "This operation may imply loss of provenance."
                print "[Enter] to continue, [Ctrl + C] to exit"
                raw_input()

                from aiida.backends.djsite.db.models import DbAttribute

                DbAttribute.set_value_for_node(code.dbnode, 'remote_exec_path', set_params.remote_abs_path)

        # store comment, to track history
        code.add_comment(comment, user=get_automatic_user())

    def code_delete(self, *args):
        """
        Delete a code

        Does not delete the code if there are calculations that are using it
        (i.e., if there are output links)
        """
        from aiida.common.exceptions import InvalidOperation
        from aiida.orm.code import delete_code

        if len(args) != 1:
            print >> sys.stderr, ("after 'code delete' there should be one "
                                  "argument only, being the code id.")
            sys.exit(1)

        code = self.get_code(args[0])
        pk = code.pk
        try:
            delete_code(code)
        except InvalidOperation as e:
            print >> sys.stderr, e.message
            sys.exit(1)

        print "Code '{}' deleted.".format(pk)



# -*- coding: utf-8 -*-

from functools import partial

from sqlalchemy import or_, and_, func, Float
from sqlalchemy.orm import defer, aliased


from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy.models.node import DbNode, DbPath, DbLink
from aiida.backends.sqlalchemy.models.group import DbGroup

from aiida.orm import DataFactory, CalculationFactory, Group, JobCalculation, load_node
from aiida.orm.data.cif import CifData
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.structure import StructureData
from aiida.orm.implementation.general.calculation.inline import InlineCalculation


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

INDEX_NAME = "dbnode_attributes_idx"

def delete_gin_index():
    delete = "DROP INDEX IF EXISTS {};".format(INDEX_NAME)
    sa.session.execute(delete)
    sa.session.commit()
    _vacuum_analyaze()

def create_gin_index():
    insert = "CREATE INDEX {} ON db_dbnode USING gin(attributes);".format(INDEX_NAME)
    sa.session.execute(insert)
    sa.session.commit()
    _vacuum_analyaze()

def _vacuum_analyaze():
    raw_conn = sa.session.connection().engine.raw_connection().connection
    isolation_level = raw_conn.isolation_level
    raw_conn.set_isolation_level(0)
    raw_conn.cursor().execute("VACUUM ANALYZE")
    raw_conn.set_isolation_level(isolation_level)

_jsonb_func_sql = """
CREATE OR REPLACE FUNCTION jsonb_merge(JSONB, JSONB)
RETURNS JSONB AS $$
WITH json_union AS (
    SELECT * FROM JSONB_EACH($1)
    UNION ALL
    SELECT * FROM JSONB_EACH($2)
) SELECT JSON_OBJECT_AGG(key, value)::JSONB FROM json_union;
$$ LANGUAGE SQL;
"""

def _add_random_number():
    """
    Add a key random to each attributes, with a random value from 0 to 100.
    """
    sa.session.execute(_jsonb_func_sql)
    sa.session.commit()

    q = ("UPDATE db_dbnode SET attributes = jsonb_merge(attributes, "
         "('{\"random\": ' || random() * 100 || '}')::jsonb)")
    sa.session.execute(q)
    sa.session.commit()


BEGIN_GROUP = 'N_elements_'
END_GROUP = '_clean_cif_primitive_dup_filtered'

def build_query_attr(filter_):
    q = DbNode.query.join(DbGroup, DbNode.dbgroups).filter(
        DbNode.attributes.has_key(filter_),
        DbGroup.name.like("{}%{}".format(BEGIN_GROUP, END_GROUP))
    )

    return lambda: q.all()

def build_query_attr_only(filter_):
    q = (sa.session.query(DbNode.attributes[filter_])
         .join(DbGroup, DbNode.dbgroups)
         .filter(DbGroup.name.like("{}%{}".format(BEGIN_GROUP, END_GROUP)))
         .filter(DbNode.attributes.has_key(filter_)))

    return lambda: q.all()

def get_farthest_cif(with_attr=False):

    nodes = ParameterData.query().with_entities('id')
    cif_type = CifData._query_type_string

    depth = (sa.session.query(DbPath.depth)
             .filter(DbPath.child_id.in_(nodes))
             .join(DbNode, DbPath.parent)
             .filter(DbNode.type.like("%{}%".format(cif_type)))
             .order_by(DbPath.depth.desc())
             .distinct()
             [0])[0]

    q = (DbPath.query.filter(DbPath.child_id.in_(nodes))
         .join(DbNode, DbPath.parent)
         .filter(DbNode.type.like("%{}%".format(cif_type)))
         .filter(DbPath.depth == depth)
         .distinct().with_entities(DbPath.id)
         )

    res = (CifData.query(children__id__in=nodes, child_paths__id__in=q)
           .distinct().order_by(DbNode.ctime))

    if not with_attr:
           res = res.options(defer(DbNode.attributes), defer(DbNode.extras))

    return res.all()

def farthest_cif_django(distinct=True):
    nodes = ParameterData.query().with_entities('id')
    cif_type = CifData._query_type_string

    depth = (sa.session.query(DbPath.depth)
             .filter(DbPath.child_id.in_(nodes))
             .join(DbNode, DbPath.parent)
             .filter(DbNode.type.like("%{}%".format(cif_type)))
             .order_by(DbPath.depth.desc()))

    if distinct:
        depth = depth.distinct()

    depth = depth[0][0]

    q = (DbPath.query.filter(DbPath.child_id.in_(nodes))
         .join(DbNode, DbPath.parent)
         .filter(DbNode.type.like("%{}%".format(cif_type)))
         .filter(DbPath.depth == depth)
         .distinct().with_entities(DbPath.id)
         )

    res = (DbNode.query.filter(DbNode.type.like("%{}%".format(cif_type)))
          .join(DbPath, DbPath.parent_id == DbNode.id)
           .filter(DbPath.child_id.in_(nodes))
           .filter(DbPath.id.in_(q))
           .options(defer(DbNode.attributes), defer(DbNode.extras))
           .distinct()
           )

    return res.all()


def get_closest_struc(with_attr=False, distinct=True):
    nodes = ParameterData.query().with_entities('id')
    struc_type = StructureData._query_type_string

    depth = (sa.session.query(DbPath.depth)
             .filter(DbPath.child_id.in_(nodes))
             .join(DbNode, DbPath.parent)
             .filter(DbNode.type.like("{}%".format(struc_type)))
             .order_by(DbPath.depth))

    if distinct:
        depth = depth.distinct()

    depth = depth[0][0]

    q = (DbPath.query.filter(DbPath.child_id.in_(nodes))
         .join(DbNode, DbPath.parent)
         .filter(DbNode.type.like("{}%".format(struc_type)))
         .filter(DbPath.depth == depth)
         )

    q = q.distinct()

    q = q.with_entities(DbPath.id)

    res = (StructureData.query()
           .join(DbPath, DbNode.child_paths)
           .filter(DbPath.child_id.in_(nodes))
           .filter(DbPath.id.in_(q))
          )

    res = res.distinct()

    res = res.order_by(DbNode.ctime)

    if not with_attr:
        res = res.options(defer(DbNode.attributes), defer(DbNode.extras))

    return res.all()

def get_closest_struc_django(distinct=True):
    nodes = ParameterData.query().with_entities('id')
    struc_type = StructureData._query_type_string

    depth = (sa.session.query(DbPath.depth)
             .filter(DbPath.child_id.in_(nodes))
             .join(DbNode, DbPath.parent)
             .filter(DbNode.type.like("{}%".format(struc_type)))
             .order_by(DbPath.depth))

    if distinct:
        depth = depth.distinct()

    depth = depth[0][0]

    q = (DbPath.query.filter(DbPath.child_id.in_(nodes))
         .join(DbNode, DbPath.parent)
         .filter(DbNode.type.like("{}%".format(struc_type)))
         .filter(DbPath.depth == depth)
         )

    q = q.distinct()

    q = q.with_entities(DbPath.id)

    res = (StructureData.query()
           .join(DbPath, DbNode.child_paths)
           .filter(DbPath.child_id.in_(nodes))
           .filter(DbPath.id.in_(q))
          )

    res = res.distinct()

    res = res.order_by(DbNode.ctime)

    if not with_attr:
        res = res.options(defer(DbNode.attributes), defer(DbNode.extras))

    return res.all()


def complex_query():
    RemoteData = DataFactory('remote')
    PwCalculation = CalculationFactory('quantumespresso.pw')

    pws = sa.session.query(DbNode.id).filter(
        DbNode.type == PwCalculation._plugin_type_string).cte()

    scratchremotes = sa.session.query(DbNode.id).filter(
        DbNode.type == RemoteData._plugin_type_string
    ).join(pws, DbNode.inputs).cte()

    inlines = sa.session.query(DbNode.id).filter(
        DbNode.type == InlineCalculation._plugin_type_string
    ).join(scratchremotes, DbNode.inputs).cte()

    storeremotes = sa.session.query(DbNode.uuid).filter(
        DbNode.type == RemoteData._plugin_type_string
    ).join(inlines, DbNode.inputs)

    return lambda: storeremotes.all()

def complex_query_no_cte():
    RemoteData = DataFactory('remote')
    PwCalculation = CalculationFactory('quantumespresso.pw')

    pws = sa.session.query(DbNode.id).filter(
        DbNode.type == PwCalculation._plugin_type_string
    )

    scratchremotes = sa.session.query(DbNode.id).filter(
        DbNode.type == RemoteData._plugin_type_string
    ).join(DbNode, DbNode.inputs, aliased=True).filter(
        DbNode.id.in_(pws)
    )

    inlines = sa.session.query(DbNode.id).filter(
        DbNode.type == InlineCalculation._plugin_type_string
    ).join(DbNode, DbNode.inputs, aliased=True).filter(
        DbNode.id.in_(scratchremotes)
    )

    storeremotes = sa.session.query(DbNode.uuid).filter(
        DbNode.type == RemoteData._plugin_type_string
    ).join(DbNode, DbNode.inputs, aliased=True).filter(
        DbNode.id.in_(inlines)
    )

    return lambda: storeremotes.all()

def complex_query_django_replicate():
    RemoteData = DataFactory('remote')
    PwCalculation = CalculationFactory('quantumespresso.pw')

    pws = sa.session.query(DbNode.id).filter(
        DbNode.type == PwCalculation._plugin_type_string
    )

    scratchremotes = sa.session.query(DbNode.id).filter(
        DbNode.type == RemoteData._plugin_type_string
    ).join(DbLink, DbNode.id == DbLink.output_id).filter(
        DbLink.input_id.in_(pws)
    )

    inlines = sa.session.query(DbNode.id).filter(
        DbNode.type == InlineCalculation._plugin_type_string
    ).join(DbLink, DbNode.id == DbLink.output_id).filter(
        DbLink.input_id.in_(scratchremotes)
    )

    storeremotes = sa.session.query(DbNode.uuid).filter(
        DbNode.type == RemoteData._plugin_type_string
    ).join(DbLink, DbNode.id == DbLink.output_id).filter(
        DbLink.input_id.in_(inlines)
    )

    return lambda: storeremotes.all()

def list_data_structure(element=None):
    q = (sa.session.query(DbNode.id, DbNode.attributes['kinds'], DbNode.attributes['sites'])
         .filter(DbNode.type.like("data.structure.%"),
                 or_(DbNode.attributes.has_key('kinds'),
                     DbNode.attributes.has_key('sites'))
                 ).order_by(DbNode.ctime))
    if element:
        d = {"kinds": [{"symbols": [element]}]}
        q = q.filter(DbNode.attributes.contains(d))

    return lambda: q.all()

def mounet1():
    pw_calc = Group.get(pk=1139193).nodes.next()
    structure = pw_calc.out.output_structure
    qstruc = StructureData.query(children__pk=structure.pk).with_entities(DbNode.id)
    n_children = aliased(DbNode)
    qic = (InlineCalculation.query()
           .join(DbLink, DbNode.id == DbLink.output_id).filter(
               DbLink.input_id.in_(qstruc)
           )
           .join(n_children, DbNode.inputs).filter(
               or_(
                   n_children.attributes["radii_source"].astext.like("%alvarez"),
                   n_children.attributes[("lowdim_dict", "radii_source")].astext.like("%alvarez")
               )
           ).distinct()
        )

    return qic.with_entities(func.count(DbNode.id)).scalar()

def mounet2():
    StructureData = DataFactory('structure')
    structure = load_node(2304207)
    qstruc = StructureData.query(children__pk=structure.pk).with_entities(DbNode.id)

    n_children = aliased(DbNode)
    qic = (InlineCalculation.query()
           .filter(
               DbNode.attributes["function_name"].astext == "lowdimfinder_inline"
           )
           .join(DbLink, DbNode.id == DbLink.output_id).filter(
               DbLink.input_id.in_(qstruc)
           )
           .join(n_children, DbNode.inputs).filter(
               or_(
                   n_children.attributes["radii_source"].astext.like("%alvarez"),
                   n_children.attributes[("lowdim_dict", "radii_source")].astext.like("%alvarez")
               )
           ).distinct()
        )

    return qic.with_entities(func.count(DbNode.id)).scalar()

def mounet_daemon():
    return JobCalculation.query(
        dbattributes__key='state', dbattributes__tval='WITHSCHEDULER'
    ).with_entities(func.count(DbNode.id)).scalar()

def range_queries(percentage):
    return sa.session.query(DbNode.attributes["random"]).filter(
        DbNode.attributes["random"].cast(Float) < percentage
    ).all()

queries = {
    "attributes": {
        'kinds': build_query_attr('kinds'),
        'sites': build_query_attr('sites'),
        'cell': build_query_attr('cell'),
    },
    "attributes_only": {
        'kinds': build_query_attr_only('kinds'),
        'sites': build_query_attr_only('sites'),
        'cell': build_query_attr_only('cell'),
    },
    "complex": {
        "1": complex_query(),
        "no_cte": complex_query_no_cte(),
        "django_replicate": complex_query_django_replicate()
    },
    "paths": {
        "farthest_cif": get_farthest_cif,
        "closest_struc": get_closest_struc,
        "closest_struc_no_distinct": partial(get_closest_struc, distinct=False),
        "farthest_cif_django": farthest_cif_django,
        "farthest_cif_django_no_distinct": partial(farthest_cif_django, distinct=False)
    },
    "paths_with_attr": {
        "farthest_cif": partial(get_farthest_cif, with_attr=True),
        "closest_struc": partial(get_closest_struc, with_attr=True),
    },
    "verdi": {
        "list_data": list_data_structure(),
        "list_element": list_data_structure(element="C")
    },
    "mounet": {
        "1": mounet1,
        "2": mounet2,
        "daemon": mounet_daemon
    }
}

queries["range"] = dict((i, partial(range_queries, i)) for i in xrange(1, 101))



# -*- coding: utf-8 -*-

"""
The dummy model encodes the model defined by django in backends.djsite
using SQLAlchemy.
This is done to query the database with more performant ORM of SA.
"""


from sqlalchemy.ext.declarative import declarative_base
from sa_init import (
    Column, Table, ForeignKey,
    Integer, String, DateTime, Float, Boolean, Text,  # basic column types
    UUID, JSONB,                                      # Fancy column types
    UniqueConstraint,aliased,
    select, func, join, and_, or_, not_, except_,     # join and filter ops
    relationship, backref, column_property,           # Table to table relationsships
    sessionmaker, create_engine,                      # connection
    foreign, mapper, case, cast
)

# Aiida Django classes:
#from aiida.orm.implementation.django.node import Node as DjangoAiidaNode

#from aiida.common.pluginloader import load_plugin


# SETTINGS:
from aiida.common.setup import get_profile_config
from aiida.backends import settings

#EXCEPTIONS
from aiida.common.exceptions import DbContentError, MissingPluginError

# MISC
from aiida.backends.sqlalchemy.models.utils import uuid_func
from aiida.utils import timezone



__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

Base = declarative_base()

class DbLink(Base):
    __tablename__ = "db_dblink"
    id = Column(Integer, primary_key=True)
    input_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
    )
    output_id = Column(
        Integer,
        ForeignKey(
            'db_dbnode.id',
            ondelete="CASCADE",
            deferrable=True,
            initially="DEFERRED"
        )
    )
    type=Column(String(255))
    input = relationship("DbNode", primaryjoin="DbLink.input_id == DbNode.id")
    output = relationship("DbNode", primaryjoin="DbLink.output_id == DbNode.id")
    label = Column(String(255), index=True, nullable=False)


class DbPath(Base):
    __tablename__ = "db_dbpath"
    id = Column(Integer, primary_key=True)
    parent_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
    )
    child_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
    )
    parent = relationship(
        "DbNode",
        primaryjoin="DbPath.parent_id == DbNode.id",
        backref="child_paths"
    )
    child = relationship(
        "DbNode",
        primaryjoin="DbPath.child_id == DbNode.id",
        backref="parent_paths"
    )
    depth = Column(Integer)
    entry_edge_id = Column(Integer)
    direct_edge_id = Column(Integer)
    exit_edge_id = Column(Integer)

class DbCalcState(Base):
    __tablename__ = "db_dbcalcstate"
    id = Column(Integer, primary_key=True)
    dbnode = relationship(
            'DbNode',
            backref=backref('dbstates', passive_deletes=True)
        )
    state = Column(String(255))
    time = Column(DateTime(timezone=True), default=timezone.now)
    dbnode_id = Column(
            Integer,
            ForeignKey(
                'db_dbnode.id', ondelete="CASCADE",
                deferrable=True, initially="DEFERRED"
            )
        )
    __table_args__ = (
        UniqueConstraint('dbnode_id', 'state'),
    )

class DbAttribute(Base):
    __tablename__ = "db_dbattribute"
    id  = Column(Integer, primary_key = True)
    dbnode_id = Column(Integer, ForeignKey('db_dbnode.id'))
    key = Column(String(255))
    datatype = Column(String(10))
    tval = Column(String, default='')
    fval = Column(Float, default=None, nullable=True)
    ival = Column(Integer, default=None, nullable=True)
    bval = Column(Boolean, default=None, nullable=True)
    dval = Column(DateTime, default=None, nullable = True)



class DbExtra(Base):
    __tablename__ = "db_dbextra"
    id  = Column(Integer, primary_key = True)
    dbnode_id = Column(Integer, ForeignKey('db_dbnode.id'))
    key = Column(String(255))
    datatype = Column(String(10))
    tval = Column(String, default='')
    fval = Column(Float, default=None, nullable=True)
    ival = Column(Integer, default=None, nullable=True)
    bval = Column(Boolean, default=None, nullable=True)
    dval = Column(DateTime, default=None, nullable = True)


class DbComputer(Base):
    __tablename__ = "db_dbcomputer"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    name = Column(String(255), unique=True, nullable=False)
    hostname = Column(String(255))

    description = Column(Text, nullable=True)
    enabled = Column(Boolean)

    transport_type = Column(String(255))
    scheduler_type = Column(String(255))

    transport_params = Column(String(255))
    _metadata = Column('metadata', String(255), default="{}")

    def get_aiida_class(self):
        from aiida.backends.djsite.db.models import DbComputer as DjangoSchemaDbComputer
        djcomputer = DjangoSchemaDbComputer(
            id=self.id, uuid=self.uuid, name=self.name,
            hostname=self.hostname, description=self.description,
            enabled=self.enabled, transport_type=self.transport_type,
            scheduler_type=self.scheduler_type, transport_params=self.transport_params,
            metadata=self._metadata
        )
        return djcomputer.get_aiida_class()

class DbUser(Base):
    __tablename__ = "db_dbuser"

    id = Column(Integer, primary_key=True)
    email = Column(String(254), unique=True, index=True)
    password = Column(String(128))  # Clear text password ?
    first_name = Column(String(254), nullable=True)
    last_name = Column(String(254), nullable=True)
    institution = Column(String(254), nullable=True)

    is_staff = Column(Boolean, default=False)
    is_active = Column(Boolean, default=False)

    last_login = Column(DateTime(timezone=True), default=timezone.now)
    date_joined = Column(DateTime(timezone=True), default=timezone.now)

    def get_aiida_class(self):
        from aiida.backends.djsite.db.models import DbUser as DjangoSchemaDbUser
        djuser = DjangoSchemaDbUser(
            id=self.id,email=self.email, password=self.password,
            first_name=self.first_name, last_name=self.last_name,
            institution=self.institution, is_staff=self.is_staff,
            is_active=self.is_active, last_login=self.last_login,
            date_joined=self.date_joined
        )
        return djuser.get_aiida_class()


table_groups_nodes = Table(
    'db_dbgroup_dbnodes',
    Base.metadata,
    Column('id', Integer, primary_key=True),
    Column(
            'dbnode_id', Integer,
            ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
        ),
    Column(
            'dbgroup_id', Integer,
            ForeignKey('db_dbgroup.id', deferrable=True, initially="DEFERRED")
        )
)

class DbGroup(Base):
    __tablename__ = "db_dbgroup"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    name = Column(String(255), index=True)

    type = Column(String(255), default="", index=True)

    time = Column(DateTime(timezone=True), default=timezone.now)
    description = Column(Text, nullable=True)

    user_id = Column(
            Integer,
            ForeignKey(
                'db_dbuser.id', ondelete='CASCADE',
                deferrable=True, initially="DEFERRED")
            )
    user = relationship('DbUser', backref=backref('dbgroups', cascade='merge'))

    dbnodes = relationship('DbNode', secondary=table_groups_nodes,
                           backref="dbgroups", lazy='dynamic')

    __table_args__ = (
        UniqueConstraint('name', 'type'),
    )

    def __str__(self):
        if self.type:
            return '<DbGroup [type: {}] "{}">'.format(self.type, self.name)
        else:
            return '<DbGroup [user-defined] "{}">'.format(self.name)
    def get_aiida_class(self):
        from aiida.orm.implementation.django.group import Group as DjangoAiidaGroup
        from aiida.backends.djsite.db.models import DbGroup as DjangoSchemaDbGroup
        dbgroup = DjangoSchemaDbGroup(
            id = self.id,
            type = self.type,
            uuid = self.uuid,
            name = self.name,
            time = self.time,
            description  = self.description,
            user_id = self.user_id,
        )

        return DjangoAiidaGroup(dbgroup=dbgroup)



class DbNode(Base):
    __tablename__ = "db_dbnode"
    id = Column(Integer, primary_key=True)
    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    type = Column(String(255), index=True)
    label = Column(String(255), index=True, nullable=True)
    description = Column(Text(), nullable=True)
    ctime = Column(DateTime(timezone=True), default=timezone.now)
    mtime = Column(DateTime(timezone=True), default=timezone.now)
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', deferrable=True, initially="DEFERRED"),
        nullable=True
    )
    dbcomputer = relationship(
        'DbComputer',
        backref=backref('dbnodes', passive_deletes=True)
    )
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', deferrable=True, initially="DEFERRED"),
        nullable=False
    )
    user = relationship('DbUser', backref='dbnodes')

    public = Column(Boolean, default=False)

    nodeversion = Column(Integer, default=1)

    attributes = relationship('DbAttribute', uselist=True, backref='dbnode')
    extras = relationship('DbExtra', uselist=True, backref='dbnode')



    outputs = relationship(
        "DbNode",
        secondary       =   "db_dblink",
        primaryjoin     =   "DbNode.id == DbLink.input_id",
        secondaryjoin   =   "DbNode.id == DbLink.output_id",
        backref         =   backref("inputs", passive_deletes=True),
        passive_deletes =   True
    )

    children = relationship(
        "DbNode",
        secondary       =   "db_dbpath",
        primaryjoin     =   "DbNode.id == DbPath.parent_id",
        secondaryjoin   =   "DbNode.id == DbPath.child_id",
        backref         =   "parents"
    )
    def get_aiida_class(self):
        """
        Return the corresponding instance of
        :func:`~aiida.orm.implementation.django.node.Node`
        or a subclass return by the plugin loader.

        .. todo::
            The behavior is quite pathetic, creating a django DbNode instance
            to instantiate the aiida instance.
            These means that every time you load Aiida instances with
            the QueryBuilder when using Django as a backend, three instances
            are instantiated for every Aiida instance you load!
            Could be fixed by allowing DbNode from the dummy nodel to be passed
            to AiidaNode's __init__.

        :returns: An instance of the plugin class
        """
        # I need to import the DbNode in the Django model,
        # and instantiate an object that has the same attributes as self.
        from aiida.backends.djsite.db.models import DbNode as DjangoSchemaDbNode
        dbnode = DjangoSchemaDbNode(
                id=self.id, type=self.type, uuid=self.uuid, ctime=self.ctime,
                mtime=self.mtime, label=self.label,
                dbcomputer_id=self.dbcomputer_id, user_id=self.user_id,
                public=self.public, nodeversion=self.nodeversion
        )
        return dbnode.get_aiida_class()




states = select(
        [
            DbCalcState.dbnode_id.label('dbnode_id'),
            func.max(DbCalcState.time).label('lasttime'),
        ]
    ).group_by(DbCalcState.dbnode_id).alias()

recent_states = select([
        DbCalcState.id.label('id'),
        DbCalcState.dbnode_id.label('dbnode_id'),
        DbCalcState.state.label('state'),
        states.c.lasttime.label('time')
    ]).\
    select_from(
        join(
            DbCalcState,
            states,
            and_(
                DbCalcState.dbnode_id == states.c.dbnode_id,
                DbCalcState.time == states.c.lasttime,
            )
        )
    ).alias() # .group_by(DbCalcState.dbnode_id, DbCalcState.time)

state_mapper = mapper(
    DbCalcState,
    recent_states,
    primary_key= recent_states.c.dbnode_id,
    non_primary=True,
)

DbNode.state_instance = relationship(
    state_mapper,
    primaryjoin = recent_states.c.dbnode_id == foreign(DbNode.id),
    viewonly=True,
)

DbNode.state = column_property(
    select([recent_states.c.state]).
    where(recent_states.c.dbnode_id == foreign(DbNode.id))
)


DbAttribute.value_str = column_property(
        case([
            (DbAttribute.datatype == 'txt', DbAttribute.tval),
            (DbAttribute.datatype == 'float', cast(DbAttribute.fval, String)),
            (DbAttribute.datatype == 'int', cast(DbAttribute.ival, String)),
            (DbAttribute.datatype == 'bool', cast(DbAttribute.bval, String)),
            (DbAttribute.datatype == 'date', cast(DbAttribute.dval, String)),
            (DbAttribute.datatype == 'txt', cast(DbAttribute.tval, String)),
            (DbAttribute.datatype == 'float', cast(DbAttribute.fval, String)),
            (DbAttribute.datatype == 'list', None),
            (DbAttribute.datatype == 'dict', None),
        ])
    )

DbAttribute.value_float = column_property(
        case([
            (DbAttribute.datatype == 'txt', cast(DbAttribute.tval, Float)),
            (DbAttribute.datatype == 'float', DbAttribute.fval),
            (DbAttribute.datatype == 'int', cast(DbAttribute.ival, Float)),
            (DbAttribute.datatype == 'bool', cast(DbAttribute.bval, Float)),
            (DbAttribute.datatype == 'date', cast(DbAttribute.dval, Float)),
            (DbAttribute.datatype == 'txt', cast(DbAttribute.tval, Float)),
            (DbAttribute.datatype == 'float', cast(DbAttribute.fval, Float)),
            (DbAttribute.datatype == 'list', None),
            (DbAttribute.datatype == 'dict', None),
        ])
    )



profile = get_profile_config(settings.AIIDADB_PROFILE)


def get_aldjemy_session():
    """
    Use aldjemy to make a session

    .. note:
        Use only in this case. In normal production mode
        it is safer make session explictly because it is more robust
    """
    from aldjemy.core import get_engine
    engine = get_engine()
    _Session = sessionmaker(bind=engine)
    return _Session()


if profile['AIIDADB_NAME'] == ':memory:':
    session = get_aldjemy_session()
else:
    engine = profile["AIIDADB_ENGINE"]
    if engine == "sqlite3":
        engine_url = (
            "sqlite:///{AIIDADB_NAME}"
            ).format(**get_profile_config(settings.AIIDADB_PROFILE))
    elif engine.startswith("mysql"):
        engine_url = (
            "mysql://{AIIDADB_USER}:{AIIDADB_PASS}@"
            "{AIIDADB_HOST}:{AIIDADB_PORT}/{AIIDADB_NAME}"
            ).format(**get_profile_config(settings.AIIDADB_PROFILE))
    elif engine.startswith("postgre"):
        engine_url = (
            "postgresql://{AIIDADB_USER}:{AIIDADB_PASS}@"
            "{AIIDADB_HOST}:{AIIDADB_PORT}/{AIIDADB_NAME}"
            ).format(**get_profile_config(settings.AIIDADB_PROFILE))
    else:
        raise ConfigurationError("Unknown DB engine: {}".format(engine))
    session = sessionmaker(bind=create_engine(engine_url))()





# -*- coding: utf-8 -*-

from aiida.parsers.parser import Parser
from aiida.common.datastructures import calc_states
from aiida.common.exceptions import PluginInternalError

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class BaseCodtoolsParser(Parser):
    """
    Base class for parsers for cod-tools package scripts.
    """

    def __init__(self, calc):
        """
        Initialize the instance of BaseCodtoolsParser
        """
        # Check for valid input:
        self._check_calc_compatibility(calc)
        super(BaseCodtoolsParser, self).__init__(calc)

    def _check_calc_compatibility(self, calc):
        from aiida.common.exceptions import ParsingError

        if not isinstance(calc, self._supported_calculation_class):
            raise ParsingError("Input calc must be a {}".format(
                self._supported_calculation_class.__name__))

    def parse_with_retrieved(self, retrieved):
        """
        Receives in input a dictionary of retrieved nodes.
        Does all the logic here.
        """
        from aiida.common.exceptions import InvalidOperation
        import os

        output_path = None
        error_path = None
        try:
            output_path, error_path = self._fetch_output_files(retrieved)
        except InvalidOperation:
            raise
        except IOError as e:
            self.logger.error(e.message)
            return False, ()

        if output_path is None and error_path is None:
            self.logger.error("No output files found")
            return False, ()

        try:
            return self._get_output_nodes(output_path, error_path)
        except PluginInternalError as e:
            self.logger.error("Internal plugin error: {}".format(e.message))
            return False, ()

    def _fetch_output_files(self, retrieved):
        """
        Checks the output folder for standard output and standard error
        files, returns their absolute paths on success.
        
        :param retrieved: A dictionary of retrieved nodes, as obtained from the
          parser.
        """
        from aiida.common.exceptions import InvalidOperation
        import os

        # check in order not to overwrite anything
        state = self._calc.get_state()
        if state != calc_states.PARSING:
            raise InvalidOperation("Calculation not in {} state"
                                   .format(calc_states.PARSING))

        # Check that the retrieved folder is there 
        try:
            out_folder = retrieved[self._calc._get_linkname_retrieved()]
        except KeyError:
            raise IOError("No retrieved folder found")

        list_of_files = out_folder.get_folder_list()

        output_path = None
        error_path = None

        if self._calc._DEFAULT_OUTPUT_FILE in list_of_files:
            output_path = os.path.join(out_folder.get_abs_path('.'),
                                       self._calc._DEFAULT_OUTPUT_FILE)
        if self._calc._DEFAULT_ERROR_FILE in list_of_files:
            error_path = os.path.join(out_folder.get_abs_path('.'),
                                      self._calc._DEFAULT_ERROR_FILE)

        return output_path, error_path

    def _get_output_nodes(self, output_path, error_path):
        """
        Extracts output nodes from the standard output and standard error
        files.
        """
        import os

        from aiida.orm.data.cif import CifData
        from aiida.orm.data.parameter import ParameterData

        cif = None
        if output_path is not None and os.path.getsize(output_path) > 0:
            cif = CifData(file=output_path)

        messages = []
        if error_path is not None:
            with open(error_path) as f:
                content = f.readlines()
            messages = [x.strip('\n') for x in content]
            self._check_failed(messages)

        output_nodes = []
        success = True
        if cif is not None:
            output_nodes.append(('cif', cif))
        else:
            success = False

        output_nodes.append(('messages',
                             ParameterData(dict={'output_messages':
                                                     messages})))

        return success, output_nodes

    def _check_failed(self, messages):
        """
        Check the STDERR for traces of typical Perl misconfiguration
        errors.

        :raises PluginInternalError: if such traces are found.
        """
        import re

        for msg in messages:
            if re.search('^(Can\'t locate|BEGIN failed)', msg):
                raise PluginInternalError(" ".join(messages))



# -*- coding: utf-8 -*-

from aiida.orm import Data
from aiida.orm.data.simple import NumericType, SimpleData, Float, Int, Str


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


_TYPE_MAPPING = {
    int: NumericType,
    float: NumericType,
    bool: SimpleData,
    str: SimpleData
}


def Bool(value):
    return SimpleData(typevalue=(bool, value))


def to_db_type(value):
    if isinstance(value, Data):
        return value
    elif isinstance(value, bool):
        return SimpleData(typevalue=(bool, value))
    elif isinstance(value, (int, long)):
        return SimpleData(typevalue=(int, value))
    elif isinstance(value, float):
        return Float(value)
    elif isinstance(value, basestring):
        return SimpleData(typevalue=(type(value), value))
    else:
        raise ValueError("Cannot convert value to database type")


def to_native_type(data):
    if not isinstance(data, Data):
        return data
    elif isinstance(data, SimpleData):
        return data.value
    else:
        raise ValueError("Cannot convert from database to native type")


def get_db_type(native_type):
    if issubclass(native_type, Data):
        return native_type
    if native_type in _TYPE_MAPPING:
        return _TYPE_MAPPING[native_type]
    return None



# -*- coding: utf-8 -*-
"""
Functions to manage the global settings stored in the DB (in the DbSettings
table.
"""

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


def set_global_setting(key, value, description=None):
    """
    Set a global setting in the DbSetting table (therefore, stored at the DB
    level).
    """
    from aiida.backends.djsite.db.models import DbSetting

    # Before storing, validate the key
    DbSetting.validate_key(key)

    # This also saves in the DB    
    DbSetting.set_value(key, value,
                        other_attribs={"description": description})


def del_global_setting(key):
    """
    Return the value of the given setting, or raise a KeyError if the
    setting is not present in the DB.
    
    :raise KeyError: if the setting does not exist in the DB
    """
    from aiida.backends.djsite.db.models import DbSetting
    from django.core.exceptions import ObjectDoesNotExist

    try:
        setting = DbSetting.objects.get(key=key)
    except ObjectDoesNotExist:
        raise KeyError("No global setting with key={}".format(key))

    # This does not raise exceptions
    DbSetting.del_value(key=key)


def get_global_setting(key):
    """
    Return the value of the given setting, or raise a KeyError if the
    setting is not present in the DB.
    
    :raise KeyError: if the setting does not exist in the DB
    """
    from aiida.backends.djsite.db.models import DbSetting
    from django.core.exceptions import ObjectDoesNotExist

    try:
        return DbSetting.objects.get(key=key).getvalue()
    except ObjectDoesNotExist:
        raise KeyError("No global setting with key={}".format(key))


def get_global_setting_description(key):
    """
    Return the description for the given setting variable, as stored in the
    DB, or raise a KeyError if the setting is not present in the DB.
    """
    from aiida.backends.djsite.db.models import DbSetting
    from django.core.exceptions import ObjectDoesNotExist

    try:
        return DbSetting.objects.get(key=key).description
    except ObjectDoesNotExist:
        raise KeyError("No global setting with key={}".format(key))



# -*- coding: utf-8 -*-
from aiida.orm.implementation.general.user import AbstractUser
from aiida.backends.sqlalchemy.models.user import DbUser
from aiida.utils.email import normalize_email

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class User(AbstractUser):

    def __init__(self, **kwargs):
        # If no arguments are passed, then create a new DbUser
        if not kwargs:
            raise ValueError("User can not be instantiated without arguments")

        # If a DbUser is passed as argument. Just use it and
        # wrap it with a User object
        elif 'dbuser' in kwargs:
            # When a dbuser is passed as argument, then no other arguments
            # should be passed.
            if len(kwargs) > 1:
                raise ValueError("When a DbUser is passed as argument, no"
                                 "further arguments are accepted.")
            dbuser = kwargs.pop('dbuser')
            if not isinstance(dbuser, DbUser):
                raise ValueError("Expected a DbUser. Object of a different"
                                 "class was given as argument.")
            self._dbuser = dbuser

        # If the email of a users is given then create a new User object with
        # this email.
        elif 'email' in kwargs:
            # When a dbuser is passed as argument, then no other arguments
            # should be passed.
            if len(kwargs) > 1:
                raise ValueError("When an email is passed as argument, no"
                                 "further arguments are accepted.")
            email = normalize_email(kwargs.pop('email'))
            self._dbuser = DbUser(email=email)

        else:
            raise ValueError("Only dbuser & email are accepted as arguments")

    @property
    def pk(self):
        return self._dbuser.id

    @property
    def id(self):
        return self._dbuser.id

    @property
    def to_be_stored(self):
        return self._dbuser.id is None

    def save(self):
        if not self.to_be_stored:
            self._dbuser.save()

    def force_save(self):
        self._dbuser.save()

    @property
    def email(self):
        return self._dbuser.email

    @email.setter
    def email(self, val):
        self._dbuser.email = val
        if not self.to_be_stored:
            self._dbuser.save()

    def _set_password(self, val):
        self._dbuser.password = val
        self.save()

    def _get_password(self):
        return self._dbuser.password

    @property
    def is_superuser(self):
        return self._dbuser.is_superuser

    @is_superuser.setter
    def is_superuser(self, val):
        self._dbuser.is_superuser = val
        self.save()

    @property
    def first_name(self):
        return self._dbuser.first_name

    @first_name.setter
    def first_name(self, val):
        self._dbuser.first_name = val
        self.save()

    @property
    def last_name(self):
        return self._dbuser.last_name

    @last_name.setter
    def last_name(self, val):
        self._dbuser.last_name = val
        self.save()

    @property
    def institution(self):
        return self._dbuser.institution

    @institution.setter
    def institution(self, val):
        self._dbuser.institution = val
        self.save()

    @property
    def is_staff(self):
        return self._dbuser.is_staff

    @is_staff.setter
    def is_staff(self, val):
        self._dbuser.is_staff = val
        self.save()

    @property
    def is_active(self):
        return self._dbuser.is_active

    @is_active.setter
    def is_active(self, val):
        self._dbuser.is_active = val
        self.save()

    @property
    def last_login(self):
        return self._dbuser.last_login

    @last_login.setter
    def last_login(self, val):
        self._dbuser.last_login = val
        self.save()

    @property
    def date_joined(self):
        return self._dbuser.date_joined

    @date_joined.setter
    def date_joined(self, val):
        self._dbuser.date_joined = val
        self.save()

    @classmethod
    def search_for_users(cls, **kwargs):

        id = kwargs.pop('id', None)
        if id is None:
            id = kwargs.pop('pk', None)
        email = kwargs.pop('email', None)

        # Constructing the default query
        dbuser_query = DbUser.query

        # If an id is specified then we add it to the query
        if id is not None:
            dbuser_query = dbuser_query.filter_by(id=id)

        # If an email is specified then we add it to the query
        if email is not None:
            dbuser_query = dbuser_query.filter_by(email=email)

        dbusers = dbuser_query.all()
        users = []
        for dbuser in dbusers:
            users.append(cls(dbuser=dbuser))
        return users



# -*- coding: utf-8 -*-

from aiida.orm.implementation.general.calculation.inline import InlineCalculation

from aiida.orm.data import Data
from aiida.common.exceptions import ModificationNotAllowed
from aiida.common.links import LinkType
from aiida.utils.calculation import add_source_info

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"


def make_inline(func):
    def wrapped_function(*args, **kwargs):
        """
        This wrapper function is the actual function that is called.
        """
        from django.db import transaction

        # Note: if you pass a lambda function, the name will be <lambda>; moreover
        # if you define a function f, and then do "h=f", h.__name__ will
        # still return 'f'!
        function_name = func.__name__
        if not function_name.endswith('_inline'):
            raise ValueError("The function name that is wrapped must end "
                             "with '_inline', while its name is '{}'".format(
                function_name))

        if args:
            print args
            raise ValueError("Arguments of inline function should be "
                             "passed as key=value")

        # Check the input values
        for k, v in kwargs.iteritems():
            if not isinstance(v, Data):
                raise TypeError("Input data to a wrapped inline calculation "
                                "must be Data nodes")
                # kwargs should always be strings, no need to check
                # if not isinstance(k, basestring):
                #    raise TypeError("")

        # Create the calculation (unstored)
        c = InlineCalculation()
        # Add data input nodes as links
        for k, v in kwargs.iteritems():
            c.add_link_from(v, label=k)

        # Try to get the source code
        add_source_info(c, func)

        # Run the wrapped function
        retval = func(**kwargs)

        # Check the output values
        if not isinstance(retval, dict):
            raise TypeError("The wrapped function did not return a dictionary")
        for k, v in retval.iteritems():
            if not isinstance(k, basestring):
                raise TypeError("One of the key of the dictionary returned by "
                                "the wrapped function is not a string: "
                                "'{}'".format(k))
            if not isinstance(v, Data):
                raise TypeError("One of the values (for key '{}') of the "
                                "dictionary returned by the wrapped function "
                                "is not a Data node".format(k))
            if v.is_stored:
                raise ModificationNotAllowed(
                    "One of the values (for key '{}') of the "
                    "dictionary returned by the wrapped function "
                    "is already stored! Note that this node (and "
                    "any other side effect of the function) are "
                    "not going to be undone!".format(k))

        # Add link to output data nodes
        for k, v in retval.iteritems():
            v.add_link_from(c, label=k, link_type=LinkType.RETURN)

        with transaction.commit_on_success():
            # I call store_all for the Inline calculation;
            # this will store also the inputs, if needed.
            c.store_all(with_transaction=False)
            # As c is already stored, I just call store (and not store_all)
            # on each output
            for v in retval.itervalues():
                v.store(with_transaction=False)

        # Return the calculation and the return values
        return c, retval

    return wrapped_function



# -*- coding: utf-8 -*-
"""
This module defines the main data structures used by the Calculation.
"""
from aiida.common.extendeddicts import DefaultFieldsAttributeDict, Enumerate

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class CalcState(Enumerate):
    pass


_sorted_datastates = (
    'NEW',  # just created
    'TOSUBMIT',  # used by the executionmanager to submit new calculations scheduled to be submitted
    'SUBMITTING',  # being submitted to cluster
    'WITHSCHEDULER',  # on the scheduler (on any unfinished status:
    # QUEUED, QUEUED_HELD, SUSPENDED, RUNNING)
    'COMPUTED',  # Calculation finished on scheduler, not yet retrieved
    # (both DONE and FAILED)
    'RETRIEVING',  # while retrieving data
    'PARSING',  # while parsing data
    'FINISHED',  # Final state of the calculation: data retrieved and eventually parsed
    'SUBMISSIONFAILED',  # error occurred during submission phase
    'RETRIEVALFAILED',  # error occurred during retrieval phase
    'PARSINGFAILED',  # error occurred during parsing phase due to a problem in the parse
    'FAILED',  # The parser recognized the calculation as failed
    'IMPORTED',  # The calculation was imported from another DB
)

# The order of states is not random: is the order of precedence.
# However, this is never used at the moment in the code.
calc_states = CalcState(_sorted_datastates)


def sort_states(list_states):
    """
    Given a list of state names, return a sorted list of states (the first
    is the most recent) sorted according to their logical appearance in
    the DB (i.e., NEW before of SUBMITTING before of FINISHED).

    .. note:: The order of the internal variable _sorted_datastates is
      used.

    :param list_states: a list (or tuple) of state strings.

    :return: a sorted list of the given data states.

    :raise ValueError: if any of the given states is not a valid state.
    """
    datastates_order_dict = {state: idx for idx, state in enumerate(
        _sorted_datastates)}

    try:
        list_to_sort = [(datastates_order_dict[st], st)
                        for st in list_states]
    except KeyError as e:
        raise ValueError("At least one of the provided states is not "
                         "valid ({})".format(e.message))

    # In-place sort
    list_to_sort.sort()

    return [_[1] for _ in list_to_sort[::-1]]


class CalcInfo(DefaultFieldsAttributeDict):
    """
    This object will store the data returned by the calculation plugin and to be
    passed to the ExecManager
    """
    # Note: some of the variables might have never been used in AiiDA
    #       one might want to clean all this stuff in a future revision
    # Note: probably some of the fields below are not used anymore inside
    #       calcinfo, but are rather directly set from calculation attributes to
    #       the JobInfo to be passed to the ExecManager
    #       (see, for instance, 'queue_name').

    _default_fields = (
        'job_environment',  # TODO UNDERSTAND THIS!
        'email',
        'email_on_started',
        'email_on_terminated',
        'uuid',
        'prepend_text',
        'append_text',
#        'cmdline_params',  # as a list of strings. These 5 variables are now in CalcInfo
#        'stdin_name',
#        'stdout_name',
#        'stderr_name',
#        'join_files',
        # 'queue_name', This is not used in CalcInfo, it is automatically set from
        # calculation attributes to JobInfo
        'num_machines',
        'num_mpiprocs_per_machine',
        'priority',
        'max_wallclock_seconds',
        'max_memory_kb',
        'rerunnable',
        'retrieve_list',  # a list of files or patterns to retrieve, with two
        # possible formats: [ 'remotepath',  # just the name of the file to retrieve. Will be put in '.' of the repositorym with name os.path.split(item)[1]
        # ['remotepath','localpath',depth]  ]
        # second format will copy the remotepath file/folder to localpath.
        # if remotepath is a file/folder, localpath will be its local name
        # if remotepath has file patterns, localpath should only be '.'
        # depth is an integer to decide the localname: will be os.path.join(localpath, filename )
        # where filename takes remotepath.split() and joins the last #depth elements
        # use the second option if you are using file patterns (*,[0-9],...)
        # ALL PATHS ARE RELATIVE!
        'local_copy_list',  # a list of length-two tuples with (localabspath, relativedestpath)
        'remote_copy_list',  # a list of length-three tuples with (remotemachinename, remoteabspath, relativedestpath)
        'remote_symlink_list',
        # a list of length-three tuples with (remotemachinename, remoteabspath, relativedestpath)
        'retrieve_singlefile_list',  # a list of files, that will be retrieved
        # from cluster and saved in SinglefileData nodes
        # in the following format:
        # ["linkname_from calc to singlefile","subclass of singlefile","filename"]
        # filename remote = filename local
        'codes_info',  # a list of dictionaries used to pass the info of the execution of a code.
        'codes_run_mode', # a string used to specify the order in which multi codes can be executed
    )


class CodeRunmode(Enumerate):
    pass

# these are the possible ways to execute more than one code in the same scheduling job
# if parallel, the codes will be executed as something like:
#   code1.x &
#   code2.x &
#   wait
# if serial, it will be:
#   code1.x
#   code2.x
code_run_modes = CodeRunmode(('PARALLEL',
                              'SERIAL'))


class CodeInfo(DefaultFieldsAttributeDict):
    """
    This attribute-dictionary contains the information needed to execute a code.
    Possible attributes are:

    * ``cmdline_params``: a list of strings, containing parameters to be written on
      the command line right after the call to the code, as for example::

        code.x cmdline_params[0] cmdline_params[1] ... < stdin > stdout

    * ``stdin_name``: (optional) the name of the standard input file. Note, it is
      only possible to use the stdin with the syntax::

        code.x < stdin_name

      If no stdin_name is specified, the string "< stdin_name" will not be
      passed to the code.
      Note: it is not possible to substitute/remove the '<' if stdin_name is specified;
      if that is needed, avoid stdin_name and use instead the cmdline_params to
      specify a suitable syntax.
    * ``stdout_name``: (optional) the name of the standard output file. Note, it is
      only possible to pass output to stdout_name with the syntax::

        code.x ... > stdout_name

      If no stdout_name is specified, the string "> stdout_name" will not be
      passed to the code.
      Note: it is not possible to substitute/remove the '>' if stdout_name is specified;
      if that is needed, avoid stdout_name and use instead the cmdline_params to
      specify a suitable syntax.
    * ``stderr_name``: (optional) a string, the name of the error file of the code.
    * ``join_files``: (optional) if True, redirects the error to the output file.
      If join_files=True, the code will be called as::

        code.x ... > stdout_name 2>&1

      otherwise, if join_files=False and stderr is passed::

        code.x ... > stdout_name 2> stderr_name

    * ``withmpi``: if True, executes the code with mpirun (or another MPI installed
      on the remote computer)
    * ``code_uuid``: the uuid of the code associated to the CodeInfo
    """
    _default_fields = ('cmdline_params',  # as a list of strings
                       'stdin_name',
                       'stdout_name',
                       'stderr_name',
                       'join_files',
                       'withmpi',
                       'code_uuid'
                       )


class WorkflowState(Enumerate):
    pass


wf_states = WorkflowState((
    'CREATED',
    'INITIALIZED',
    'RUNNING',
    'FINISHED',
    'SLEEP',
    'ERROR'
))


class WorkflowDataType(Enumerate):
    pass


wf_data_types = WorkflowDataType((
    'PARAMETER',
    'RESULT',
    'ATTRIBUTE',
))


class WorkflowDataValueType(Enumerate):
    pass


wf_data_value_types = WorkflowDataValueType((
    'NONE',
    'JSON',
    'AIIDA',
))

wf_start_call = "start"
wf_exit_call = "exit"
wf_default_call = "none"

# TODO Improve/implement this!
# class DynResourcesInfo(AttributeDict):
#    """
#    This object will contain a list of 'dynamical' resources to be
#    passed from the code plugin to the ExecManager, containing
#    things like
#    * resources in the permanent repository, that will be simply
#      linked locally (but copied remotely on the remote computer)
#      to avoid a waste of permanent repository space
#    * remote resources to be directly copied over only remotely
#    """
#    pass
